# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2016)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines the `Tab` API, and all of the built-in tab objects
"""

# core
from .registry import (
    register_tab,
    get_tab,
)
from .core import (
    _MetaTab,
    BaseTab,
    StaticTab,
    GpsTab,
    IntervalTab,
    EventTab,
    Tab,
    ParentTab,
    TabList,
)
from .builtin import (
    ExternalTab,
    PlotTab,
    StateTab,
    UrlTab,
)
from .misc import (
    AboutTab,
    Error404Tab,
)

# data
from .data import (ProcessedTab, DataTab)

# application-specific extras
from .sei import SEIWatchDogTab
from .guardian import GuardianTab
from .stamp import StampPEMTab
from .management import AccountingTab
from .etg import EventTriggerTab
from .fscan import FscanTab
from .gracedb import GraceDbTab

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Pipeline generator for the Gravitational-wave interferometer
summary information system (`gwsumm`)

This module constructs a directed acyclic graph (DAG) that defines
a workflow to be submitted via the HTCondor scheduler.
"""

import argparse
import os
import shutil
import sys

from glue import pipeline

from gwdatafind.utils import find_credential

from gwpy.io import kerberos as gwkerberos

from gwdetchar import cli

from . import __version__
from .utils import mkdir

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'

PROG = ('python -m gwsumm.batch' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))


# -- utilities ----------------------------------------------------------------

class GWSummaryJob(pipeline.CondorDAGJob):
    """Job representing a configurable instance of gw_summary.
    """
    logtag = '$(cluster)-$(process)'

    def __init__(self, universe, tag='gw_summary',
                 subdir=None, logdir=None, **cmds):
        pipeline.CondorDAGJob.__init__(self, universe, sys.executable)
        if subdir:
            subdir = os.path.abspath(subdir)
            self.set_sub_file(os.path.join(subdir, '%s.sub' % (tag)))
        if logdir:
            logdir = os.path.abspath(logdir)
            self.set_log_file(os.path.join(
                logdir, '%s-%s.log' % (tag, self.logtag)))
            self.set_stderr_file(os.path.join(
                logdir, '%s-%s.err' % (tag, self.logtag)))
            self.set_stdout_file(os.path.join(
                logdir, '%s-%s.out' % (tag, self.logtag)))
        cmds.setdefault('getenv', 'True')
        for key, val in cmds.items():
            if hasattr(self, 'set_%s' % key.lower()):
                getattr(self, 'set_%s' % key.lower())(val)
            else:
                self.add_condor_cmd(key, val)
        # add python module sub-command
        self._command = ' '.join(['-m', __package__])

    def add_opt(self, opt, value=''):
        pipeline.CondorDAGJob.add_opt(self, opt, str(value))
    add_opt.__doc__ = pipeline.CondorDAGJob.add_opt.__doc__

    def set_command(self, command):
        self._command = ' '.join([
            self._command,
            command,
        ])

    def get_command(self):
        return self._command

    def write_sub_file(self):
        pipeline.CondorDAGJob.write_sub_file(self)
        # insert positional arguments in the right place
        with open(self.get_sub_file(), 'r') as f:
            sub = f.read()
        sub = sub.replace(
            'arguments = "',
            'arguments = " {0}'.format(self.get_command()),
        )
        with open(self.get_sub_file(), 'w') as f:
            f.write(sub)


class GWSummaryDAGNode(pipeline.CondorDAGNode):
    def get_cmd_line(self):
        # merge positional arguments with options
        return ' '.join([
            self.job().get_command(),
            pipeline.CondorDAGNode.get_cmd_line(self),
        ])


# -- parse command-line -------------------------------------------------------

class GWHelpFormatter(argparse.HelpFormatter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('indent_increment', 4)
        super(GWHelpFormatter, self).__init__(*args, **kwargs)


def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    usage = ('%(prog)s --global-config defaults.ini --config-file '
             'myconfig.ini [--config-file myconfig2.ini] [options]')
    parser = argparse.ArgumentParser(
        prog=PROG,
        usage=usage,
        description=__doc__,
        formatter_class=GWHelpFormatter,
    )
    bopts = parser.add_argument_group("Basic options")
    htcopts = parser.add_argument_group("Condor options")
    copts = parser.add_argument_group(
        "Configuration options",
        "Each --global-config file will be used in all nodes of the workflow, "
        "while a single node will be created for each other --config-file",
    )
    popts = parser.add_argument_group(
        "Process options",
        "Configure how this summary will be processed.",
    )
    outopts = parser.add_argument_group("Output options")
    topts = parser.add_argument_group(
        "Time mode options",
        "Choose a stadard time mode, or a GPS [start, stop) interval",
    )

    # general arguments
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="show verbose output, default: %(default)s",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="show program's version number and exit",
    )
    parser.version = __version__

    # basic options
    bopts.add_argument(
        '-i',
        '--ifo',
        action='store',
        type=str,
        metavar='IFO',
        help="Instrument to process. If this option is set in "
             "the [DEFAULT] of any of the INI files, giving it "
             "here is redundant.",
    )
    wrapgroup = bopts.add_mutually_exclusive_group()
    wrapgroup.add_argument(
        '-w',
        '--skip-html-wrapper',
        action='store_true',
        default=False,
        help="Do not configure first job for HTML htmlnode, default: "
             "%(default)s. Useful for separating large summary pipeline "
             "across multiple DAGs",
    )
    wrapgroup.add_argument(
        '-W',
        '--html-wrapper-only',
        action='store_true',
        help="Only run first job for HTML htmlnode.",
    )
    bopts.add_argument(
        '-t',
        '--file-tag',
        action='store',
        type=str,
        default='gw_summary_pipe',
        help="file tag for pipeline files, default: %(default)s",
    )

    # HTCondor options
    htcopts.add_argument(
        '-u',
        '--universe',
        action='store',
        type=str,
        default='vanilla',
        help="Universe for condor jobs, default: %(default)s",
    )
    htcopts.add_argument(
        '-l',
        '--log-dir',
        action='store',
        type=str,
        default=os.environ.get('LOCALDIR', None),
        help="Directory path for condor log files, default: %(default)s",
    )
    htcopts.add_argument(
        '-m',
        '--maxjobs',
        action='store',
        type=int,
        default=None,
        metavar='N',
        help="Restrict the DAG to submit only N jobs at any one "
             "time, default: %(default)s",
    )
    htcopts.add_argument(
        '-T',
        '--condor-timeout',
        action='store',
        type=float,
        default=None,
        metavar='T',
        help='Configure condor to terminate jobs after T hours '
             'to prevent idling, default: %(default)s',
    )
    htcopts.add_argument(
        '-c',
        '--condor-command',
        action='append',
        type=str,
        default=[],
        help="Extra condor submit commands to add to gw_summary submit file. "
             "Can be given multiple times in the form \"key=value\"",
    )

    # configuration options
    copts.add_argument(
        '-f',
        '--config-file',
        action='append',
        type=str,
        metavar='FILE',
        default=[],
        help="INI file for analysis, may be given multiple times",
    )
    copts.add_argument(
        '-g',
        '--global-config',
        action='append',
        type=str,
        metavar='FILE',
        default=[],
        help="INI file for use in all workflow jobs, may be given "
             "multiple times",
    )
    copts.add_argument(
        '-p',
        '--priority',
        action='append',
        type=str,
        default=[],
        help="priority for DAG node, should be given once "
             "for each --config-file in the same order",
    )

    # process options
    popts.add_argument(
        '--nds',
        action='store_true',
        default='guess',
        help='use NDS as the data source, default: %(default)s',
    )
    popts.add_argument(
        '--single-process',
        action='store_true',
        default=False,
        help="restrict gw_summary to a single process, mainly for "
             "debugging purposes, default: %(default)s",
    )
    popts.add_argument(
        '--multi-process',
        action='store',
        type=int,
        default=None,
        help="maximum number of concurrent sub-processes for each "
             "gw_summary job, {number of CPUs} / {min(number of jobs, 4)}",
    )
    popts.add_argument(
        '-a',
        '--archive',
        action='store_true',
        default=False,
        help="Read archived data from the FILE, and "
             "write back to it at the end",
    )
    popts.add_argument(
        '-S',
        '--on-segdb-error',
        action='store',
        type=str,
        default='raise',
        choices=['raise', 'ignore', 'warn'],
        help="action upon error fetching segments from SegDB",
    )
    popts.add_argument(
        '-G',
        '--on-datafind-error',
        action='store',
        type=str,
        default='raise',
        choices=['raise', 'ignore', 'warn'],
        help="action upon error querying for frames from the "
             "datafind server: default: %(default)s",
    )
    popts.add_argument(
        '--data-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of TimeSeries data files',
    )
    popts.add_argument(
        '--event-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of event trigger files',
    )
    popts.add_argument(
        '--segment-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of state '
             'or data-quality segment files',
    )
    popts.add_argument(
        '--no-htaccess',
        action='store_true',
        default=False,
        help='tell gw_summary to not write .htaccess files',
    )

    # output options
    outopts.add_argument(
        '-o',
        '--output-dir',
        action='store',
        type=str,
        metavar='OUTDIR',
        default=os.curdir,
        help="Output directory for summary information, "
             "default: '%(default)s'",
    )

    # time mode options
    topts.add_argument(
        "--day",
        action="store",
        type=str,
        metavar='YYYYMMDD',
        help="UTC date to process",
    )
    topts.add_argument(
        "--week",
        action="store",
        type=str,
        metavar="YYYYMMDD",
        help="week to process (by UTC starting date)",
    )
    topts.add_argument(
        "--month",
        action="store",
        type=str,
        metavar="YYYYMM",
        help="calendar month to process",
    )
    topts.add_argument(
        "--year",
        action="store",
        type=str,
        metavar="YYYY",
        help="calendar year to process",
    )
    topts.add_argument(
        "-s",
        "--gps-start-time",
        action="store",
        type=int,
        metavar="GPSSTART",
        help="GPS start time",
    )
    topts.add_argument(
        "-e",
        "--gps-end-time",
        action="store",
        type=int,
        metavar="GPSEND",
        help="GPS end time",
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the command-line Omega scan tool in batch mode
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    # initialize logger
    logger = cli.logger(
        name=PROG.split('python -m ').pop(),
        level='DEBUG' if args.verbose else 'INFO',
    )

    # check time options
    N = sum([args.day is not None, args.month is not None,
             args.gps_start_time is not None, args.gps_end_time is not None])
    if N > 1 and not (args.gps_start_time and args.gps_end_time):
        raise parser.error("Please give only one of --day, --month, or "
                           "--gps-start-time and --gps-end-time.")

    for (i, cf) in enumerate(args.config_file):
        args.config_file[i] = ','.join(map(os.path.abspath, cf.split(',')))
    args.global_config = list(map(
        os.path.abspath,
        [fp for csv in args.global_config for fp in csv.split(',')],
    ))

    # -- build workflow directories -----------------

    # move to output directory
    indir = os.getcwd()
    mkdir(args.output_dir)
    os.chdir(args.output_dir)
    outdir = os.curdir

    # set node log path, and condor log path
    logdir = os.path.join(outdir, 'logs')
    htclogdir = args.log_dir or logdir
    mkdir(logdir, htclogdir)

    # set config directory and copy config files
    etcdir = os.path.join(outdir, 'etc')
    mkdir(etcdir)

    for (i, fp) in enumerate(args.global_config):
        inicopy = os.path.join(etcdir, os.path.basename(fp))
        if not os.path.isfile(inicopy) or not os.path.samefile(fp, inicopy):
            shutil.copyfile(fp, inicopy)
        args.global_config[i] = os.path.abspath(inicopy)
    for (i, csv) in enumerate(args.config_file):
        inicopy = []
        for fp in csv.split(','):
            fp2 = os.path.join(etcdir, os.path.basename(fp))
            if not os.path.isfile(fp2) or not os.path.samefile(fp, fp2):
                shutil.copyfile(fp, fp2)
            inicopy.append(os.path.abspath(fp2))
        args.config_file[i] = ','.join(inicopy)
    logger.debug("Copied all INI configuration files to %s" % etcdir)

    # -- configure X509 and kerberos for condor -----

    if args.universe != 'local':
        # copy X509 grid certificate into local location
        (x509cert, _) = find_credential()
        x509copy = os.path.join(etcdir, os.path.basename(x509cert))
        shutil.copyfile(x509cert, x509copy)

        # rerun kerberos with new path
        krb5cc = os.path.abspath(os.path.join(etcdir, 'krb5cc.krb5'))
        gwkerberos.kinit(krb5ccname=krb5cc)
        logger.debug("Configured Condor and Kerberos "
                     "for NFS-shared credentials")

    # -- build DAG ----------------------------------

    dag = pipeline.CondorDAG(os.path.join(htclogdir, '%s.log' % args.file_tag))
    dag.set_dag_file(os.path.join(outdir, args.file_tag))

    universe = args.universe

    # -- parse condor commands ----------------------

    # parse into a dict
    condorcmds = {}
    if args.condor_timeout:
        condorcmds['periodic_remove'] = (
            'CurrentTime-EnteredCurrentStatus > %d' %
            (3600 * args.condor_timeout)
        )
    for cmd_ in args.condor_command:
        (key, value) = cmd_.split('=', 1)
        condorcmds[key.rstrip().lower()] = value.strip()

    if args.universe != 'local':
        # add X509 to environment
        for (env_, val_) in zip(['X509_USER_PROXY', 'KRB5CCNAME'],
                                [os.path.abspath(x509copy), krb5cc]):
            condorenv = '%s=%s' % (env_, val_)
            if ('environment' in condorcmds and
                    env_ not in condorcmds['environment']):
                condorcmds['environment'] += ';%s' % condorenv
            elif 'environment' not in condorcmds:
                condorcmds['environment'] = condorenv

    # -- build individual gw_summary jobs -----------

    globalconfig = ','.join(args.global_config)

    jobs = []
    if not args.skip_html_wrapper:
        htmljob = GWSummaryJob(
            'local', subdir=outdir, logdir=logdir,
            tag='%s_local' % args.file_tag, **condorcmds)
        jobs.append(htmljob)
    if not args.html_wrapper_only:
        datajob = GWSummaryJob(
            universe, subdir=outdir, logdir=logdir,
            tag=args.file_tag, **condorcmds)
        jobs.append(datajob)

    # add common command-line options
    for job in jobs:
        if args.day:
            job.set_command('day')
            job.add_arg(args.day)
        elif args.week:
            job.set_command('week')
            job.add_arg(args.week)
        elif args.month:
            job.set_command('month')
            job.add_arg(args.month)
        elif args.year:
            job.set_command('year')
            job.add_arg(args.year)
        elif args.gps_start_time or args.gps_end_time:
            job.set_command('gps')
            job.add_arg(str(args.gps_start_time))
            job.add_arg(str(args.gps_end_time))
        else:
            job.set_command('day')
        if args.nds is True:
            job.add_opt('nds')
        if args.single_process:
            job.add_opt('single-process')
        elif args.multi_process is not None:
            job.add_opt('multi-process', args.multi_process)
        if args.verbose:
            job.add_opt('verbose')
        if args.ifo:
            job.add_opt('ifo', args.ifo)
        job.add_opt('on-segdb-error', args.on_segdb_error)
        job.add_opt('on-datafind-error', args.on_datafind_error)
        job.add_opt('output-dir', outdir)
        for (opt, fplist) in zip(
                ['--data-cache', '--event-cache', '--segment-cache'],
                [args.data_cache, args.event_cache, args.segment_cache]):
            if fplist:
                job.add_arg('%s %s' % (opt, (' %s ' % opt).join(fplist)))
        if args.no_htaccess:
            job.add_opt('no-htaccess')

    # make surrounding HTML first
    if not args.skip_html_wrapper:
        htmljob.add_opt('html-only', '')
        htmljob.add_opt('config-file', ','.join(
            [globalconfig]+args.config_file).strip(','))

        htmlnode = GWSummaryDAGNode(htmljob)
        for configfile in args.config_file:
            htmlnode.add_input_file(args.config_file)
        htmlnode.set_category('gw_summary')
        dag.add_node(htmlnode)
        logger.debug(" -- Configured HTML htmlnode job")

    # create node for each config file
    if not args.html_wrapper_only:
        # add html opts
        datajob.add_opt('no-html', '')
        if args.archive:
            datajob.add_condor_cmd('+SummaryNodeType', '"$(macroarchive)"')
        # configure each data node
        for (i, configfile) in enumerate(args.config_file):
            node = GWSummaryDAGNode(datajob)
            node.add_var_arg('--config-file %s' % ','.join(
                [globalconfig, configfile]).strip(','))
            if args.archive:
                jobtag = os.path.splitext(os.path.basename(configfile))[0]
                archivetag = jobtag.upper().replace('-', '_')
                if args.ifo and archivetag.startswith('%s_' %
                                                      args.ifo.upper()):
                    archivetag = archivetag[3:]
                node.add_var_opt('archive', archivetag)
            for cf in configfile.split(','):
                node.add_input_file(cf)
            node.set_category('gw_summary')
            try:
                node.set_priority(args.priority[i])
            except IndexError:
                node.set_priority(0)
            node.set_retry(1)
            if not args.skip_html_wrapper:
                node.add_parent(htmlnode)
            dag.add_node(node)
            logger.debug(" -- Configured job for config %s" % configfile)

    if args.maxjobs:
        dag.add_maxjobs_category('gw_summary', args.maxjobs)

    # -- finish up ----------------------------------

    dag.write_sub_files()
    dag.write_dag()
    dag.write_script()
    logger.debug("Setup complete, DAG written to: {}".format(
            os.path.abspath(dag.get_dag_file())))

    # return to original directory
    os.chdir(indir)


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Registry for `states <SummaryState>`.
"""

from .. import globalv
from ..utils import re_quote

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

__all__ = ['register_state', 'get_state', 'get_states']


def register_state(state, key=None, force=False):
    """Register a new `SummaryState` to the given ``key``

    Parameters
    ----------
    state : `SummaryState`
        defining Class for this state type.
    key : `str`, optional
        unique descriptive name for the `SummaryState` to be registered.
        If ``key=None``, the :attr:`~SummaryState.key`
        attribute of the given state will be used.
    force : `bool`
        overwrite existing registration for this key

    Raises
    ------
    ValueError
        if key is already registered and ``force`` not given as `True`
    """
    if key is None:
        key = state.key
    key = key.lower()
    if key not in globalv.STATES or force:
        globalv.STATES[key] = state
        return state
    raise ValueError("State %r has already been registered." % key)


def get_state(key):
    """Query the registry for the `SummaryState` registered to the given key

    Parameters
    ----------
    key : `str`
        registered key of desired `SummaryState`. This may not match the
        `~SummaryState.name` attribute` if the state was registered with
        a different key.

    Returns
    -------
    state : `SummaryState`
        the `SummaryState` registered with the given key

    Raises
    ------
    ValueError:
        if the ``key`` doesn't map to a registered `SummaryState`
    """
    key = re_quote.sub('', key)
    try:
        return globalv.STATES[key.lower()]
    except KeyError:
        raise ValueError("No SummaryState registered with name '%s'" % key)


def get_states(keys=set()):
    """Query the registry for a list of states (defaults to all)

    Parameters
    ----------
    keys : `set` of `str`
        the set of state keys to query in the registry

    Returns
    -------
    states : `dict`
        a `dict` of (``key``, `SummaryState`) pairs

    Raises
    ------
    ValueError:
        if any of the ``keys`` doesn't map to a registered `SummaryState`
    """
    if not keys:
        return globalv.STATES.copy()
    else:
        return dict((key, get_state(key)) for key in keys)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwsumm.tabs`

"""

import os.path

import pytest

from .. import tabs
from ..plot import SummaryPlot

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- gwpy.tabs.registry -------------------------------------------------------

@pytest.mark.parametrize('name, tab', [
    ('basic', tabs.Tab),
    ('\'plots\'', tabs.PlotTab),
])
def test_get_tab(name, tab):
    assert tabs.get_tab(name) is tab
    with pytest.raises(ValueError):
        tabs.get_tab('fasmklwea')


def test_register_tab():
    class TestTab(object):
        type = 'test'
        pass

    tabs.register_tab(TestTab)
    with pytest.raises(ValueError):
        tabs.register_tab(TestTab)
    tabs.register_tab(TestTab, force=True)

    assert tabs.get_tab('test') is TestTab

    tabs.register_tab(TestTab, name='test-with-name')
    assert tabs.get_tab('test-with-name') is TestTab


# -- test tab classes ---------------------------------------------------------

class TestTab(object):
    TYPE = 'basic'
    DEFAULT_ARGS = ['Test']

    @classmethod
    def setup_class(cls):
        cls.TAB = tabs.get_tab(cls.TYPE)

    def create(self, *args, **kwargs):
        args = list(args)
        while len(args) < len(self.DEFAULT_ARGS):
            args.append(self.DEFAULT_ARGS[len(args)])
        return self.TAB(*args, **kwargs)

    def test_init(self):
        self._test_init('Test')

    def _test_init(self, *args, **kwargs):
        if len(args) == 0:
            args = self.DEFAULT_ARGS
        # test basic creation and defaults
        tab = self.create(*args, **kwargs)
        assert tab.type == self.TYPE
        assert tab.name == args[0]
        assert tab.shortname == kwargs.pop('shortname', tab.name)
        assert tab.children == kwargs.pop('children', [])
        assert tab.parent == kwargs.pop('parent', None)
        assert tab.group == kwargs.pop('group', None)
        assert tab.path == kwargs.pop('path', os.path.curdir)
        assert tab.hidden == kwargs.pop('hidden', False)
        return tab

    def test_shortname(self):
        tab = self.create()
        assert tab.shortname == tab.name
        tab = self.create('Test', shortname='ShortTest')
        assert tab.shortname == 'ShortTest'

    def test_index(self):
        tab = self.create()
        assert tab.index == os.path.join('test', 'index.html')
        tab2 = self.create('Parent')
        del tab.index
        tab.set_parent(tab2)
        assert tab.index == os.path.join('parent', 'test', 'index.html')


# -- external tab

class TestExternalTab(TestTab):
    TYPE = 'external'
    DEFAULT_ARGS = ['Test', '//test.com']

    def test_init(self):
        tab = self._test_init()
        assert tab.url == '//test.com'


# -- plot tab

class TestPlotTab(TestTab):
    TYPE = 'plots'

    def test_init(self):
        plots = ['test.png']
        tab = self._test_init('Test', plots=plots)
        assert tab.plots == list(map(SummaryPlot, plots))
        assert tab.layout is None

    def test_add_plot(self):
        tab = self.create()
        before = tab.plots[:]
        plot = 'test.png'
        tab.add_plot(plot)
        assert tab.plots == before + [SummaryPlot(href=plot)]

    def test_layout(self):
        tab = self.create()
        tab.set_layout(1)
        assert tab.layout == [1]

        tab.set_layout((1, 2))
        assert tab.layout == [1, 2]

        tab.set_layout((1, (1, 2)))
        assert tab.layout == [1, (1, 2)]

        with pytest.raises(ValueError):
            tab.set_layout('test')
        with pytest.raises(ValueError):
            tab.set_layout([1, (1, 2, 1)])
        with pytest.warns(DeprecationWarning):
            tab.layout = [1]



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for data loading and pre-processing
"""

from collections import OrderedDict
from functools import wraps

from ligo.segments import segmentlist as LigoSegmentList

from gwpy.segments import (DataQualityFlag, SegmentList, Segment)

from ..channels import get_channel
from ..config import GWSummConfigParser

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- method decorators --------------------------------------------------------

def use_segmentlist(f):
    """Decorate a method to convert incoming segments into a `SegmentList`

    This assumes that the method to be decorated takes a segment list as
    the second positionsl argument.
    """
    @wraps(f)
    def decorated_func(arg1, segments, *args, **kwargs):
        if isinstance(segments, DataQualityFlag):
            segments = segments.active
        elif not isinstance(segments, LigoSegmentList):
            segments = SegmentList([Segment(*x) for x in segments])
        return f(arg1, segments, *args, **kwargs)
    return decorated_func


def use_configparser(f):
    """Decorate a method to use a valid default for 'config'

    This is just to allow lazy passing of `config=None`
    """
    @wraps(f)
    def decorated_func(*args, **kwargs):
        if kwargs.get('config', None) is None:
            kwargs['config'] = GWSummConfigParser()
        return f(*args, **kwargs)
    return decorated_func


# -- handle keys for globalv dicts --------------------------------------------
# need a key that is unique across channel(s) with a specific for of
# signal-processing parameters

FFT_PARAMS = OrderedDict([
    ('method', str),  # keep this one first (DMM)
    ('fftlength', float),
    ('overlap', float),
    ('window', None),
    ('stride', float),
    ('scheme', None),
])

DEFAULT_FFT_PARAMS = {
    'method': 'median',
}


class FftParams(object):
    """Convenience object to hold signal-processing parameters
    """
    __slots__ = list(FFT_PARAMS)

    def __init__(self, **kwargs):
        for slot in self.__slots__:
            kwargs.setdefault(slot, None)
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __setattr__(self, key, val):
        if val is not None and FFT_PARAMS[key] is not None:  # type cast
            val = FFT_PARAMS[key](val)
        super(FftParams, self).__setattr__(key, val)

    def __str__(self):
        out = []
        for slot in self.__slots__:
            attr = getattr(self, slot)
            if not attr:
                out.append('')
            elif slot == 'scheme':
                out.append(type(attr).__name__)
            else:
                out.append(str(attr))
        return ';'.join(out)

    def dict(self):
        return dict((x, getattr(self, x)) for x in self.__slots__ if
                    getattr(self, x) is not None)


def get_fftparams(channel, **defaults):
    # configure FftParams
    params = {k: v for k, v in DEFAULT_FFT_PARAMS.items() if v is not None}
    params.update(defaults)
    fftparams = FftParams(**params)

    # update FftParams with per-channel overrides
    channel = get_channel(channel)
    for key in fftparams.__slots__:
        try:
            setattr(fftparams, key, getattr(channel, key))
        except AttributeError:
            try:  # set attribute in channel object for future reference
                setattr(channel, key, defaults[key])
            except KeyError:
                pass

    # set stride to something sensible
    if fftparams.stride is None and fftparams.overlap:
        fftparams.stride = fftparams.fftlength * 1.5
    elif fftparams.stride is None:
        fftparams.stride = fftparams.fftlength

    # sanity check parameters
    if fftparams.fftlength == 0:
        raise ZeroDivisionError("Cannot operate with FFT length of 0")
    if fftparams.stride == 0:
        raise ZeroDivisionError("Cannot generate spectrogram with stride "
                                "length of 0")

    return fftparams


def make_globalv_key(channels, fftparams=None):
    """Generate a unique key for storing data in a globalv `dict`

    Parameters
    ----------
    channels : `str`, `list`
        one or more channels to group in this key
    fftparams : `FftParams`
        structured set of signal-processing parameters used to generate the
        dataset
    """
    if not isinstance(channels, (list, tuple)):
        channels = [channels]
    channels = list(map(get_channel, channels))
    parts = []
    # comma-separated list of names
    parts.append(','.join(c.ndsname for c in channels))
    # colon-separated list of FFT parameters
    if fftparams is not None:
        parts.append(fftparams)
    return ';'.join(map(str, parts))



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2016)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""HTML <head> helphers

This module mainly declares the resources used by standard on HTML pages
"""

from collections import OrderedDict

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


# build collection of CSS resources
CSS = OrderedDict((
    ('font-awesome', 'https://cdnjs.cloudflare.com/ajax/libs/'
                     'font-awesome/5.15.1/css/fontawesome.min.css'),
    ('font-awesome-solid', 'https://cdnjs.cloudflare.com/ajax/libs/'
                           'font-awesome/5.15.1/css/solid.min.css'),
    ('gwbootstrap', 'https://cdn.jsdelivr.net/npm/gwbootstrap@1.3.2/'
                    'lib/gwbootstrap.min.css'),
))

# build collection of javascript resources
JS = OrderedDict((
    ('jquery', 'https://code.jquery.com/jquery-3.5.1.min.js'),
    ('jquery-ui', 'https://code.jquery.com/ui/1.12.1/jquery-ui.min.js'),
    ('moment', 'https://cdnjs.cloudflare.com/ajax/libs/'
               'moment.js/2.29.1/moment.min.js'),
    ('bootstrap', 'https://cdn.jsdelivr.net/npm/bootstrap@4.5.3/'
                  'dist/js/bootstrap.bundle.min.js'),
    ('fancybox', 'https://cdnjs.cloudflare.com/ajax/libs/'
                 'fancybox/3.5.7/jquery.fancybox.min.js'),
    ('datepicker', 'https://cdnjs.cloudflare.com/ajax/libs/'
                   'bootstrap-datepicker/1.9.0/js/'
                   'bootstrap-datepicker.min.js'),
    ('gwbootstrap', 'https://cdn.jsdelivr.net/npm/gwbootstrap@1.3.2/'
                    'lib/gwbootstrap-extra.min.js'),
))


# -- utilities ----------------------------------------------------------------

def get_css():
    """Return a `dict` of CSS files to link in the HTML <head>
    """
    return CSS


def get_js():
    """Return a `dict` of javascript files to link in the HTML <head>
    """
    return JS



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Definitions for event trigger plots
"""

import re
from collections import OrderedDict
from itertools import cycle

from numpy import isinf

from astropy.units import Quantity

from gwpy.detector import (Channel, ChannelList)
from gwpy.plot.gps import GPSTransform
from gwpy.plot.utils import (color_cycle, marker_cycle)
from gwpy.segments import SegmentList

from gwdetchar.plot import texify

from ... import globalv
from ...channels import get_channel
from ...data import (get_timeseries, add_timeseries)
from ..registry import (get_plot, register_plot)
from ...triggers import (get_triggers, get_time_column)
from ...utils import re_cchar
from ..utils import (get_column_string, hash)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TimeSeriesDataPlot = get_plot('timeseries')

LATEX_OPERATOR = {
    '>=': r'\geq',
    '<=': r'\leq',
}


class TriggerPlotMixin(object):
    """Mixin to overwrite `channels` property for trigger plots

    We don't need to get channel data for trigger plots.
    """
    def __init__(self, *args, **kwargs):
        self.filterstr = kwargs.pop('filterstr', None)
        super(TriggerPlotMixin, self).__init__(*args, **kwargs)

    @property
    def allchannels(self):
        """List of all unique channels for this plot
        """
        chans = set([re.split(r'[#@]', str(c), 1)[0] for c in self._channels])
        return ChannelList(map(Channel, chans))

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            chans = "".join(map(str, self.channels))
            filts = "".join(map(str, [
                getattr(c, 'filter', getattr(c, 'frequency_response', ''))
                for c in self.channels]))
            if self.filterstr:
                filts += self.filterstr
            self._pid = hash(chans + filts)
            return self.pid


class TriggerDataPlot(TriggerPlotMixin, TimeSeriesDataPlot):
    """Standard event trigger plot
    """
    type = 'triggers'
    data = 'triggers'
    defaults = TimeSeriesDataPlot.defaults.copy()
    defaults.update({
        'x': 'time',
        'y': 'snr',
        'color': None,
        'edgecolor': 'face',
        'facecolor': None,
        'marker': 'o',
        's': 20,
        'vmin': None,
        'vmax': None,
        'clim': None,
        'cmap': 'YlGnBu',
        'logcolor': False,
        'colorlabel': None,
    })

    def __init__(self, channels, start, end, state=None, outdir='.',
                 etg=None, **kwargs):
        if (len(channels) == 1) and ('title' not in kwargs):
            kwargs['title'] = texify('%s (%s)' % (str(channels[0]), etg))
        super(TriggerDataPlot, self).__init__(channels, start, end,
                                              state=state, outdir=outdir,
                                              **kwargs)
        self.etg = etg
        self.columns = [self.pargs.pop(c) for c in ('x', 'y', 'color')]

    @property
    def pid(self):
        """Unique identifier for this `TriggerDataPlot`.

        Extends the standard `TimeSeriesDataPlot` pid with the ETG
        and each of the column names.
        """
        try:
            return self._pid
        except AttributeError:
            super(TriggerDataPlot, self).pid
            self._pid += '_%s' % re_cchar.sub('_', self.etg)
            for column in self.columns:
                if column:
                    self._pid += '_%s' % re_cchar.sub('_', str(column))
            self._pid = self._pid.upper()
            return self.pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    def draw(self):
        # get columns
        xcolumn, ycolumn, ccolumn = self.columns

        # initialise figure
        plot = self.init_plot()
        ax = plot.gca()
        ax.grid(True, which='both')

        # work out labels
        labels = self.pargs.pop('labels', self.channels)
        if isinstance(labels, str):
            labels = labels.split(',')
        labels = [str(s).strip('\n ') for s in labels]

        # get colouring params
        cmap = self.pargs.pop('cmap')
        clim = self.pargs.pop('clim', self.pargs.pop('colorlim', None))
        cnorm = 'log' if self.pargs.pop('logcolor', False) else None
        clabel = self.pargs.pop('colorlabel', None)
        no_loudest = self.pargs.pop('no-loudest', False) is not False
        loudest_by = self.pargs.pop('loudest-by', None)

        # get plot arguments
        plotargs = []
        for i in range(len(self.channels)):
            plotargs.append(dict())
        # get plot arguments
        for key in ['vmin', 'vmax', 'edgecolor', 'facecolor', 'cmap', 's',
                    'marker', 'rasterized', 'sortbycolor']:
            try:
                val = self.pargs.pop(key)
            except KeyError:
                continue
            if key == 'facecolor' and len(self.channels) > 1 and val is None:
                val = color_cycle()
            if key == 'marker' and len(self.channels) > 1 and val is None:
                val = marker_cycle()
            elif isinstance(val, (list, tuple, cycle)):
                val = cycle(val)
            else:
                val = cycle([val] * len(self.channels))
            for i in range(len(self.channels)):
                plotargs[i][key] = next(val)

        # add data
        valid = SegmentList([self.span])
        if self.state and not self.all_data:
            valid &= self.state.active
        ntrigs = 0
        for channel, label, pargs in zip(self.channels, labels, plotargs):
            try:
                channel = get_channel(channel)
            except ValueError:
                pass
            if '#' in str(channel) or '@' in str(channel):
                key = '%s,%s' % (str(channel),
                                 self.state and str(self.state) or 'All')
            else:
                key = str(channel)
            table = get_triggers(key, self.etg, valid, query=False)
            if self.filterstr is not None:
                table = table.filter(self.filterstr)
            ntrigs += len(table)
            # access channel parameters for limits
            for c, column in zip(('x', 'y', 'c'), (xcolumn, ycolumn, ccolumn)):
                if not column:
                    continue
                # hack for SnglBurst frequency nonsense
                if column in ['peak_frequency', 'central_freq']:
                    column = 'frequency'
                # set x and y in plotargs
                param = '%s_range' % column
                lim = '%slim' % c
                if (getattr(channel, param, None) is not None
                        and c in ('x', 'y')):
                    self.pargs.setdefault(lim, getattr(channel, param))
                    if isinstance(self.pargs[lim], Quantity):
                        self.pargs[lim] = self.pargs[lim].value
                # set clim separately
                elif hasattr(channel, param):
                    if not clim:
                        clim = getattr(channel, param)

            ax.scatter(table[xcolumn], table[ycolumn],
                       c=table[ccolumn] if ccolumn else None,
                       label=label, **pargs)

        # customise plot
        legendargs = self.parse_legend_kwargs(markerscale=3)
        if len(self.channels) == 1:
            self.pargs.setdefault('title', texify(
                '%s (%s)' % (str(self.channels[0]), self.etg)))
        for axis in ('x', 'y'):  # prevent zeros on log scale
            scale = getattr(ax, 'get_{0}scale'.format(axis))()
            lim = getattr(ax, 'get_{0}lim'.format(axis))()
            if scale == 'log' and lim[0] <= 0 and not ntrigs:
                getattr(ax, 'set_{0}lim'.format(axis))(1, 10)

        self.apply_parameters(ax, **self.pargs)

        # correct log-scale empty axes
        if any(map(isinf, ax.get_ylim())):
            ax.set_ylim(0.1, 10)

        # add colorbar
        if ccolumn:
            if not ntrigs:
                ax.scatter([1], [1], c=[1], visible=False)
            ax.colorbar(cmap=cmap, clim=clim, norm=cnorm, label=clabel)

        if len(self.channels) == 1 and len(table) and not no_loudest:
            columns = [x for x in
                       (loudest_by or ccolumn or ycolumn, xcolumn, ycolumn,
                        ccolumn) if x is not None]
            self.add_loudest_event(ax, table, *columns, fontsize='large')

        if len(self.channels) > 1:
            ax.legend(**legendargs)

        # add state segments
        if isinstance(ax.xaxis.get_transform(), GPSTransform):
            self.add_state_segments(ax)
            self.add_future_shade()

        # finalise
        return self.finalize()

    def add_loudest_event(self, ax, table, rank, *columns, **kwargs):
        # get loudest row
        idx = table[rank].argmax()
        row = table[idx]
        x = float(row[columns[0]])
        y = float(row[columns[1]])

        # clip loudest event to axes limits
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        x1 = max(min(x, xlim[1]), xlim[0])
        y1 = max(min(y, ylim[1]), ylim[0])
        if x1 != x or y1 != y:  # loudest event is out of view
            facecolor = 'pink'
            clipon = False
        else:
            facecolor = 'gold'
            clipon = True

        # mark loudest row with star
        coll = ax.scatter([x1], [y1], marker='*', zorder=1000,
                          facecolor=facecolor, edgecolor='black',
                          s=200, clip_on=clipon)

        # get text
        txt = []
        for col in OrderedDict.fromkeys((rank,) + columns):  # unique ordered
            # format column name
            colstr = get_column_string(col)
            # format row value
            try:
                valstr = f"{row[col]:.2f}".rstrip('.0')
            except ValueError:  # not float()able
                valstr = str(row[col])
            txt.append(f'{colstr} = {valstr}')

        # get position for new text
        try:
            pos = kwargs.pop('position')
        except KeyError:  # user didn't specify, set default above title
            tpos = ax.title.get_position()
            pos = [tpos[0], tpos[1]+0.09]

        # parse text kwargs
        text_kw = {  # defaults
            'transform': ax.transAxes,
            'verticalalignment': 'bottom',
            'horizontalalignment': 'center',
        }
        text_kw.update(kwargs)
        if 'ha' in text_kw:  # handle short versions or alignment params
            text_kw['horizontalalignment'] = text_kw.pop('ha')
        if 'va' in text_kw:
            text_kw['verticalalignment'] = text_kw.pop('va')

        # add text
        text = ax.text(pos[0], pos[1],
                       f"Loudest event: {', '.join(txt)}",
                       **text_kw)

        return coll, text


register_plot(TriggerDataPlot)


class TriggerTimeSeriesDataPlot(TimeSeriesDataPlot):
    """Custom time-series plot to handle discontiguous `TimeSeries`.
    """
    type = 'trigger-timeseries'
    data = 'triggers'

    def draw(self):
        """Read in all necessary data, and generate the figure.
        """
        plot = self.init_plot()
        ax = plot.gca()

        # work out labels
        labels = self.pargs.pop('labels', self.channels)
        if isinstance(labels, str):
            labels = labels.split(',')
        labels = [str(s).strip('\n ') for s in labels]

        # add data
        for label, channel in zip(labels, self.channels):
            label = texify(label)
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            data = get_timeseries(channel, valid, query=False)
            # handle no timeseries
            if not len(data):
                ax.plot([0], [0], visible=False, label=label)
                continue
            # plot time-series
            color = None
            for ts in data:
                # double-check log scales
                if self.logy:
                    ts.value[ts.value == 0] = 1e-100
                if color is None:
                    line = ax.plot(ts, label=label)[0]
                    color = line.get_color()
                else:
                    ax.plot(ts, color=color, label=None)

            # allow channel data to set parameters
            if hasattr(data[0].channel, 'amplitude_range'):
                self.pargs.setdefault('ylim',
                                      data[0].channel.amplitude_range)

        # add horizontal lines to add
        for yval in self.pargs['hline']:
            try:
                yval = float(yval)
            except ValueError:
                continue
            else:
                ax.plot([self.start, self.end], [yval, yval],
                        linestyle='--', color='red')

        # customise plot
        legendargs = self.parse_legend_kwargs()
        self.apply_parameters(ax, **self.pargs)
        if len(self.channels) > 1:
            ax.legend(**legendargs)

        # finalise
        self.add_state_segments(ax)
        return self.finalize()


register_plot(TriggerTimeSeriesDataPlot)


class TriggerHistogramPlot(TriggerPlotMixin, get_plot('histogram')):
    """HistogramPlot from a LIGO_LW Table
    """
    type = 'trigger-histogram'
    data = 'triggers'

    def __init__(self, *args, **kwargs):
        super(TriggerHistogramPlot, self).__init__(*args, **kwargs)
        self.etg = self.pargs.pop('etg')
        self.column = self.pargs.pop('column')

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            etg = re_cchar.sub('_', self.etg).upper()
            self._pid = '%s_%s' % (etg, super(TriggerHistogramPlot, self).pid)
            if self.column:
                self._pid += '_%s' % re_cchar.sub('_', self.column).upper()
            return self.pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    def draw(self):
        """Get data and generate the figure.
        """
        # get histogram parameters
        plot = self.init_plot()
        ax = plot.gca()

        # extract histogram arguments
        histargs = self.parse_plot_kwargs()
        legendargs = self.parse_legend_kwargs()

        # add data
        data = []
        livetime = []
        for channel in self.channels:
            try:
                channel = get_channel(channel)
            except ValueError:
                pass
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            if '#' in str(channel) or '@' in str(channel):
                key = '%s,%s' % (str(channel),
                                 self.state and str(self.state) or 'All')
            else:
                key = str(channel)
            table_ = get_triggers(key, self.etg, valid, query=False)
            if self.filterstr is not None:
                table_ = table_.filter(self.filterstr)
            livetime.append(float(abs(table_.meta['segments'])))
            data.append(table_[self.column])
            # allow channel data to set parameters
            if hasattr(channel, 'amplitude_range'):
                self.pargs.setdefault('xlim', channel.amplitude_range)

        # plot
        for arr, d, pargs in zip(data, livetime, histargs):
            # set range if not given
            if pargs.get('range') is None:
                pargs['range'] = self._get_range(
                    d,
                    # use range from first dataset if already calculated
                    range=histargs[0].get('range'),
                    # use xlim if manually set (user or INI)
                    xlim=None if ax.get_autoscalex_on() else ax.get_xlim(),
                )
            pargs.setdefault('label', None)
            if pargs.get('log', True):
                pargs.setdefault('bottom', 1e-200)
            ax.hist(arr, **pargs)

        # tight scale the axes
        try:
            d = pargs.pop('orientation', 'vertical')
        except NameError:
            pass
        else:
            if d == 'vertical':
                ax.autoscale_view(tight=True, scaley=False)
            elif d == 'horizontal':
                ax.autoscale_view(tight=True, scalex=False)

        # customise plot
        self.apply_parameters(ax, **self.pargs)
        if len(self.channels) > 1:
            ax.legend(**legendargs)

        # finalise
        return self.finalize()


register_plot(TriggerHistogramPlot)


class TriggerRateDataPlot(TriggerPlotMixin, TimeSeriesDataPlot):
    """TimeSeriesDataPlot of trigger rate.
    """
    type = 'trigger-rate'
    data = 'triggers'
    defaults = TimeSeriesDataPlot.defaults.copy()
    defaults.update({
        'column': None,
        'legend-bbox_to_anchor': (1., 1.),
        'legend-loc': 'upper left',
        'legend-markerscale': 3,
        'legend-frameon': False,
        'ylabel': 'Rate [Hz]',
    })

    def __init__(self, *args, **kwargs):
        if 'stride' not in kwargs:
            raise ValueError("'stride' must be configured for all rate plots.")
        if 'column' in kwargs and 'bins' not in kwargs:
            raise ValueError("'bins' must be configured for rate plots if "
                             "'column' is given.")
        super(TriggerRateDataPlot, self).__init__(*args, **kwargs)
        self.etg = self.pargs.pop('etg')
        self.column = self.pargs.pop('column')

    @property
    def pid(self):
        try:
            return self._pid
        except AttributeError:
            etg = re_cchar.sub('_', self.etg).upper()
            pid = '%s_%s' % (etg, super(TriggerRateDataPlot, self).pid)
            if self.column:
                self.pid += '_%s' % re_cchar.sub('_', self.column).upper()
            return pid

    @pid.setter
    def pid(self, id_):
        self._pid = str(id_)

    def draw(self):
        """Read in all necessary data, and generate the figure.
        """

        # get rate arguments
        stride = self.pargs.pop('stride')
        if self.column:
            cname = get_column_string(self.column)
            bins = self.pargs.pop('bins')
            operator = self.pargs.pop('operator', '>=')
            try:
                opstr = LATEX_OPERATOR[operator]
            except KeyError:
                opstr = str(operator)
        else:
            bins = ['_']

        # work out labels
        labels = self.pargs.pop('labels', None)
        if isinstance(labels, str):
            labels = labels.split(',')
        elif labels is None and self.column and len(self.channels) > 1:
            labels = []
            for channel, bin_ in [(c, b) for c in self.channels for b in bins]:
                labels.append(r' '.join([channel, '$%s$' % opstr,
                                         str(bin_)]))
            self.pargs.setdefault('legend-title', cname)
        elif labels is None and self.column:
            labels = [r' '.join(['$%s$' % opstr, str(b)]) for b in bins]
            self.pargs.setdefault('legend-title', cname)
        elif labels is None:
            labels = self.channels
        self.pargs['labels'] = [str(s).strip('\n ') for s in labels]

        # get time column
        tcol = self.pargs.pop('timecolumn', None)

        # generate data
        keys = []
        for channel in self.channels:
            if self.state and not self.all_data:
                valid = self.state.active
            else:
                valid = SegmentList([self.span])
            if '#' in str(channel) or '@' in str(channel):
                key = '%s,%s' % (str(channel),
                                 str(self.state) if self.state else 'All')
            else:
                key = str(channel)
            table_ = get_triggers(key, self.etg, valid, query=False)
            if self.filterstr is not None:
                table_ = table_.filter(self.filterstr)
            tcol_ = tcol or get_time_column(table_, self.etg)
            if self.column:
                rates = list(table_.binned_event_rates(
                    stride, self.column, bins, operator=operator,
                    start=self.start, end=self.end, timecolumn=tcol_).values())
            else:
                rates = [table_.event_rate(stride, start=self.start,
                                           end=self.end, timecolumn=tcol_)]
            for bin, rate in zip(bins, rates):
                rate.channel = channel
                keys.append('%s_%s_EVENT_RATE_%s_%s'
                            % (str(channel), str(self.etg),
                               str(self.column), bin))
                if keys[-1] not in globalv.DATA:
                    add_timeseries(rate, keys[-1])

        # reset channel lists and generate time-series plot
        channels = self.channels
        outputfile = self.outputfile
        self.channels = keys
        out = super(TriggerRateDataPlot, self).draw(outputfile=outputfile)
        self.channels = channels
        return out


register_plot(TriggerRateDataPlot)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2018)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Plot the triggers for a given ETG and a given channel
"""

import argparse
import os
import sys

from collections import OrderedDict
from numpy import ndarray

from glue.lal import Cache

from gwpy.segments import Segment
from gwpy.time import to_gps

from gwdetchar.cli import logger

from ...segments import get_segments
from ...triggers import (get_triggers, keep_in_segments)
from ...utils import safe_eval

# set matplotlib backend
from matplotlib import use
use('Agg')

# backend-dependent imports
from matplotlib.artist import setp  # noqa: E402
from matplotlib.colors import LogNorm  # noqa: E402
from ...plot import get_plot  # noqa: E402

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'

PROG = ('python -m gwsumm.plot.triggers' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = logger(name=PROG.split('python -m ').pop())

# default columns for plot
DEFAULT_COLUMNS = ['time', 'frequency', 'snr']


# -- parse command-line -------------------------------------------------------

def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize argument parser
    parser = argparse.ArgumentParser(
        prog=PROG,
        description=__doc__,
    )
    trigopts = parser.add_argument_group('trigger options')
    popts = parser.add_argument_group('plot options')

    # required arguments
    parser.add_argument('channel')
    parser.add_argument('gpsstart', type=to_gps)
    parser.add_argument('gpsend', type=to_gps)

    # options for reading triggers
    trigopts.add_argument(
        '-e',
        '--etg',
        default='omicron',
        help="name of ETG (default: %(default)s)",
    )
    trigopts.add_argument(
        '-l',
        '--cache-file',
        help="cache file containing event trigger file references",
    )
    trigopts.add_argument(
        '-C',
        '--columns',
        type=lambda x: x.split(','),
        help="comma-separated list of columns to read from "
             "files (default: {0})".format(','.join(DEFAULT_COLUMNS)),
    )
    trigopts.add_argument(
        '-s',
        '--snr',
        default=0,
        type=float,
        help="minimum SNR (default: %(default)s)",
    )
    trigopts.add_argument(
        '-a',
        '--state',
        metavar='FLAG',
        help="restrict triggers to active times for flag",
    )

    # options for output plot
    popts.add_argument(
        '-t',
        '--epoch',
        type=to_gps,
        help="Zero-time for plot (default: gpsstart)",
    )
    popts.add_argument(
        '-x',
        '--x-column',
        help="column for X-axis (default: first --column)",
    )
    popts.add_argument(
        '-f',
        '--y-column',
        help="column for Y-axis (default: second --column)",
    )
    popts.add_argument(
        '-c',
        '--color',
        help="column for colour (default: third --column)",
    )
    popts.add_argument(
        '-p',
        '--plot-params',
        action='append',
        metavar='"{arg}={param}"',
        help="extra plotting keyword arguments, "
             "can be given multiple times",
    )
    popts.add_argument(
        '--tiles',
        action='store_true',
        default=False,
        help="plot tiles instead of dots (default: %(default)s), "
             "if selected 'duration' and 'bandwidth' will be "
             "automatically added to --columns",
    )
    popts.add_argument(
        '--state-label',
        help="label for state segments (default: --state)",
    )
    popts.add_argument(
        '-o',
        '--output-file',
        default='triggers.png',
        help="output file name (default: %(default)s)",
    )

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the online trigger visualization tool
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    args.epoch = args.epoch or args.gpsstart
    args.columns = args.columns or DEFAULT_COLUMNS
    if len(args.columns) < 2:
        parser.error("--columns must receive at least two columns, "
                     "got {0}".format(len(args.columns)))
    args.x_column = args.x_column or args.columns[0]
    args.y_column = args.y_column or args.columns[1]
    if not args.color and len(args.columns) >= 3:
        args.color = args.columns[2]

    # add columns for tile plot
    for c in ('duration', 'bandwidth'):
        if args.tiles and c not in args.columns:
            args.columns.append(c)

    span = Segment(args.gpsstart, args.gpsend)

    # format default params
    params = {
        'xscale': 'auto-gps',
        'epoch': args.epoch or args.gpsstart,
        'xlim': (args.gpsstart, args.gpsend),
        'yscale': 'log',
        'ylabel': 'Frequency [Hz]',
        'cmap': get_plot('triggers').defaults.get('cmap', 'YlGnBu'),
        'clim': (3, 50),
        'colorlabel': 'Signal-to-noise ratio (SNR)',
        'sortbycolor': True,
    }

    # update with user params
    for input_ in args.plot_params or []:
        key, val = input_.split('=', 1)
        params[key.strip('-')] = safe_eval(val)

    # -- load triggers ------------------------------

    # get segments
    if args.state:
        segs = get_segments(args.state, [span])

    # read cache
    if args.cache_file:
        with open(args.cache_file, 'r') as f:
            cache = Cache.fromfile(f).sieve(segment=span)
        LOGGER.info("Read cache of {0} files".format(len(cache)))
    else:
        cache = None

    # get triggers
    trigs = get_triggers(
        args.channel, args.etg, [span], cache=cache, columns=args.columns,
        verbose='Reading {0.channel} ({0.etg}) events'.format(args))
    LOGGER.info("Read {0} events".format(len(trigs)))
    if args.state:
        trigs = keep_in_segments(trigs, segs.active, etg=args.etg)
        LOGGER.info("{0} events in state {1!r}".format(len(trigs), args.state))
    if args.snr:
        trigs = trigs[trigs['snr'] >= args.snr]
        LOGGER.info("{0} events remaining with snr >= {1.snr}".format(
            len(trigs), args))

    # -- make plot ----------------------------------

    # format keywords for plot creation
    plot_kw = OrderedDict(
        (key, params.pop(key))
        for key in (
            'xscale',
            'xlim',
            'epoch',
            'yscale',
            'ylabel',
            'sortbycolor',
        )
    )

    # create plot
    if args.tiles:
        plot = trigs.tile(args.x_column, args.y_column,
                          'duration', 'bandwidth', color=args.color,
                          edgecolor=params.pop('edgecolor', 'face'),
                          linewidth=params.pop('linewidth', .8), **plot_kw)
    else:
        plot = trigs.scatter(args.x_column, args.y_column, color=args.color,
                             edgecolor=params.pop('edgecolor', 'none'),
                             s=params.pop('s', 12), **plot_kw)
    ax = plot.gca()
    mappable = ax.collections[0]

    # set mappable properties
    vmin, vmax = params.pop('clim', (3, 50))
    if params.pop('logcolor', True):
        mappable.set_norm(LogNorm(vmin=vmin, vmax=vmax))
    if mappable._A is None:
        mappable._A = ndarray((0,))

    # draw colorbar
    clabel = params.pop(
        'colorlabel',
        'Signal-to-noise ratio (SNR)',
    )
    cmap = params.pop(
        'cmap',
        get_plot('triggers').defaults.get('cmap',  'YlGnBu'),
    )
    ax.colorbar(mappable=mappable, label=clabel, cmap=cmap, clim=(vmin, vmax))

    for key, val in params.items():
        try:
            getattr(ax, 'set_%s' % key)(val)
        except AttributeError:
            setattr(ax, key, val)

    # add segments
    if args.state:
        sax = plot.add_segments_bar(segs, ax=ax,
                                    label=args.state_label or args.state)
        setp(sax.get_yticklabels(), fontsize=10)
        sax.set_epoch(args.epoch)

    # save and exit
    plot.save(args.output_file)
    LOGGER.info("Plot saved to {0.output_file}".format(args))


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of the 'All' state.

This is a special `SummaryState` that has valid and active segments spanning
the full analysis interval.
"""

from gwpy.segments import (Segment, SegmentList)

from ..globalv import NOW
from .core import SummaryState
from .registry import register_state

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

ALLSTATE = 'all'


def generate_all_state(start, end, register=True, **kwargs):
    """Build a new `SummaryState` for the given [start, end) interval.

    Parameters
    ----------
    start : `~gwpy.time.LIGOTimeGPS`, float
        the GPS start time of the current analysis
    end : `~gwpy.time.LIGOTimeGPS`, float
        the GPS end time of the current analysis
    register : `bool`, optional
        should the new `SummaryState` be registered, default `True`
    **kwargs
        other keyword arguments passed to the `SummaryState` constructor

    Returns
    -------
    allstate : `SummaryState`
        the newly created 'All' `SummaryState`
    """
    now = min(end, NOW)
    all_ = SummaryState(ALLSTATE,
                        known=SegmentList([Segment(start, end)]),
                        active=SegmentList([Segment(start, now)]),
                        **kwargs)
    all_.ready = True
    if register:
        register_state(all_)
    return all_



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines tabs for generating plots from data on-the-fly.

This module also provides the `ProcessedTab` mixin, which should be used
to declare that a tab has a `process()` method that should be executed
as part of a workflow, see the ``gw_summary`` executable as an example.
"""

import re
import os.path
import getpass
from configparser import (
    ConfigParser,
    NoOptionError,
    NoSectionError,
)
from copy import copy
from io import StringIO
from datetime import timedelta

from MarkupPy import markup

from numpy import isclose

from astropy.time import Time

from matplotlib import rcParams

from gwpy.segments import (Segment, SegmentList, DataQualityFlag)
from gwpy.time import from_gps
from gwpy.utils.mp import multiprocess_with_queues

from gwdetchar.io import html as gwhtml

from .. import (globalv, html)
from ..channels import (re_channel, get_channel,
                        split_combination as split_channel_combination)
from ..config import GWSummConfigParser
from ..mode import (Mode, get_mode)
from ..data import (get_timeseries_dict, get_spectrograms,
                    get_coherence_spectrograms, get_spectrum)
from ..data.timeseries import FRAMETYPE_REGEX
from ..data.utils import get_fftparams
from ..plot import get_plot
from ..segments import get_segments
from ..state import (generate_all_state, ALLSTATE, get_state)
from ..triggers import get_triggers
from ..utils import (re_flagdiv, vprint, safe_eval)

from .registry import (get_tab, register_tab)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['ProcessedTab', 'DataTab']

ParentTab = get_tab('state')


# -- ProcessTab mixin ---------------------------------------------------------

class ProcessedTab(object):
    """Abstract base class to detect necessity to run Tab.process()
    """
    type = '_processed'

    def process(self):
        """This method must be overridden by all subclasses
        """
        raise NotImplementedError("process() must be defined in %s"
                                  % type(self).__name__)


register_tab(ProcessedTab)


# -- DataTab ------------------------------------------------------------------

class DataTab(ProcessedTab, ParentTab):
    """A tab where plots and data summaries are built upon request

    This is the 'default' tab for the command-line gw_summary executable.

    All ``*args`` and ``**kwargs`` are passed up-stream to the base
    class constructor, excepting the following:

    Parameters
    ----------
    name : `str`
        name of this tab (required)
    start : `LIGOTimeGPS`, `str`
        start time of this `DataTab`, anything that can be parsed by
        `~gwpy.time.to_gps` is fine
    end : `LIGOTimeGPS`, `str`
        end time of this `DataTab`, format as for `start`
    states : `list` of `states <gwsumm.state.SummaryState>`
        the `list` of states (`~gwsumm.state.SummaryState`) over which
        this `DataTab` should be processed. More states can be added
        later (but before running :meth:`~DataTab.process`) via
        :meth:`~DataTab.add_state`.
    ismeta : `bool`, optional, default: `False`
        indicates that this tab only contains data already by others
        and so doesn't need to be processed.
    noplots : `bool`, optional, default: `False`
        indicates that this tab only exists to trigger data access, and
        shouldn't actually generate any figures
    **kwargs
        other keyword arguments

    See Also
    --------
    gwsumm.tabs.StateTab
        for details on the other keyword arguments (``**kwargs``)
        accepted by the constructor for the `DataTab`.
    """
    type = 'data'

    def __init__(self, name, states=list([ALLSTATE]), ismeta=False,
                 noplots=False, **kwargs):
        """Initialise a new `DataTab`.
        """
        super(DataTab, self).__init__(name, states=states, **kwargs)
        self.ismeta = ismeta
        self.noplots = noplots
        self.subplots = []

    # -------------------------------------------
    # SummaryTab configuration parser

    @classmethod
    def from_ini(cls, cp, section, plotdir='plots', **kwargs):
        """Define a new `SummaryTab` from the given section of the
        `ConfigParser`.

        Parameters
        ----------
        cp : :class:`~gwsumm.config.GWConfigParser`
            customised configuration parser containing given section
        section : `str`
            name of section to parse
        plotdir : `str`, optional, default: ``'plots'``
            output path for plots, relative to current directory

        Returns
        -------
        tab : `DataTab`
            a new `DataTab` defined from the configuration
        """
        kwargs.setdefault('plots', [])

        # get meta tags
        try:
            ismeta = cp.get(section, 'meta-tab')
        except NoOptionError:
            pass
        else:
            if ismeta is None:
                kwargs.setdefault('ismeta', True)
            else:
                kwargs.setdefault('ismeta', bool(ismeta.title()))
        try:
            noplots = cp.get(section, 'no-plots')
        except NoOptionError:
            pass
        else:
            if noplots is None:
                kwargs.setdefault('noplots', True)
            else:
                kwargs.setdefault('noplots', bool(noplots.title()))

        job = super(DataTab, cls).from_ini(cp, section, **kwargs)
        job._config = cp._sections[section]

        # -------------------
        # parse plot requests
        #    All config entries whose key is a single integer is
        #    interpreted as a requested plot.

        start, end = job.span

        # parse subplot request
        try:
            subidx = cp.getint(section, 'subplot')
        except NoOptionError:
            subidx = None
        else:
            job.subplots = []
            subplots = []
            try:
                subdelta = timedelta(seconds=cp.getfloat(
                    section, 'subplot-duration'))
            except NoOptionError:
                mode = get_mode()
                if mode == Mode.day:
                    subdelta = timedelta(hours=1)
                elif mode == Mode.week:
                    subdelta = timedelta(days=1)
                elif mode == Mode.month:
                    subdelta = timedelta(weeks=1)
                elif mode == Mode.year:
                    subdelta = timedelta(months=1)
                else:
                    d = int(end - start)
                    if d <= 601:
                        subdelta = timedelta(minutes=1)
                    elif d <= 7201:
                        subdelta = timedelta(minutes=10)
                    elif d <= 86401:
                        subdelta = timedelta(hours=1)
                    elif d <= 259201:
                        subdelta = timedelta(hours=6)
                    else:
                        subdelta = timedelta(days=1)
            startd = Time(float(start), format='gps', scale='utc').datetime
            endd = Time(float(end), format='gps', scale='utc').datetime
            while startd < endd:
                e = min(endd, startd + subdelta)
                sps = int(Time(startd, format='datetime', scale='utc').gps)
                spe = int(Time(e, format='datetime', scale='utc').gps)
                subplots.append((sps, spe))
                startd += subdelta

        # find and order the plots
        requests = sorted([(int(opt), val) for (opt, val) in
                           cp.nditems(section) if opt.isdigit()],
                          key=lambda a: a[0])

        # parse plot definition
        for index, definition in requests:
            # find plot customisations within this section
            mods = {}
            for key, val in cp.nditems(section):
                if key.startswith('%d-' % index):
                    mods[key.split('-', 1)[1]] = safe_eval(val)

            # parse definition for section references
            try:
                pdef, sources = [s[::-1] for s in
                                 re.split(r'[\s,]', definition[::-1], 1)]
            except ValueError:
                pdef = definition
                sources = []
            else:
                if not re_channel.match(sources) and cp.has_section(sources):
                    try:
                        sources = cp.get(sources, 'channels')
                    except NoOptionError:
                        pass

            # if pdef refers to another config section, it must have a type
            if cp.has_section(pdef):
                type_ = cp.get(pdef, 'type')
                PlotClass = get_plot(type_)
            elif (pdef not in ['range-histogram', 'segment-histogram'] and
                    pdef.endswith('-histogram')):
                type_ = None
                etg, column = pdef.rsplit('-', 2)[:2]
                mods.setdefault('etg', etg)
                mods.setdefault('column', column)
                PlotClass = get_plot('trigger-histogram')
            elif re.search(r'-rate', pdef):
                type_ = None
                etg = pdef.rsplit('-', 1)[0]
                mods.setdefault('etg', etg)
                PlotClass = get_plot('trigger-rate')
            else:
                type_ = None
                PlotClass = get_plot(pdef)
            # if the plot definition declares multiple states
            if mods.pop('all-states', False):
                mods.setdefault('all-data', True)
                if type_:
                    plot = PlotClass.from_ini(cp, pdef, start, end, sources,
                                              state=None, outdir=plotdir,
                                              **mods)
                else:
                    plot = PlotClass(sources, start, end, state=None,
                                     outdir=plotdir, **mods)
                job.plots.append(plot)
                if subidx == index:
                    for span in subplots:
                        subplot = copy(plot)
                        subplot.pargs = plot.pargs.copy()
                        subplot.span = span
                        job.subplots.append(subplot)
            # otherwise define individually for multiple states
            else:
                for state in job.states:
                    if type_:
                        plot = PlotClass.from_ini(cp, pdef, start, end,
                                                  sources, state=state,
                                                  outdir=plotdir, **mods)
                    else:
                        plot = PlotClass(sources, start, end, state=state,
                                         outdir=plotdir, **mods)
                    job.plots.append(plot)
                    if subidx == index:
                        for span in subplots:
                            subplot = copy(plot)
                            subplot.pargs = plot.pargs.copy()
                            subplot.span = span
                            job.subplots.append(subplot)

        return job

    # -------------------------------------------
    # SummaryTab processing

    def finalize_states(self, config=ConfigParser(), segdb_error='raise',
                        **kwargs):
        """Fetch the segments for each state for this `SummaryTab`
        """
        # finalize all-state
        try:
            allstate = get_state(ALLSTATE)
        except ValueError:
            allstate = generate_all_state(self.start, self.end)
        allstate.fetch(config=config, segdb_error=segdb_error, **kwargs)
        for state in self.states:
            state.fetch(config=config, segdb_error=segdb_error, **kwargs)

    def process(self, config=ConfigParser(), nproc=1, **stateargs):
        """Process data for this tab

        Parameters
        ----------
        config : `ConfigParser.ConfigParser`, optional
            job configuration to pass to :math:`~DataTab.finalize_states`
        **stateargs
            all other keyword arguments are passed directly onto the
            :meth:`~DataTab.process_state` method.
        """
        if self.ismeta:
            return
        config = GWSummConfigParser.from_configparser(config)
        # load state segments
        self.finalize_states(
            config=config, datacache=stateargs.get('datacache', None),
            segmentcache=stateargs.get('segmentcache', None),
            segdb_error=stateargs.get('segdb_error', 'raise'),
            datafind_error=stateargs.get('datafind_error', 'raise'),
            nproc=nproc, nds=stateargs.get('nds', None))
        vprint("States finalised [%d total]\n" % len(self.states))
        for state in self.states:
            vprint("    {0.name}: {1} segments | {2} seconds".format(
                state, len(state.active), abs(state.active)))
            if state is self.defaultstate:
                vprint(" [DEFAULT]")
            vprint('\n')

        # pre-process requests for 'all-data' plots
        all_data = any([(p.all_data & p.new) for p in self.plots])
        if all_data:
            vprint("Pre-processing all-data requests:\n")
            self.process_state(None, config=config, nproc=nproc,
                               **stateargs)
        # process each state
        for state in sorted(self.states, key=lambda s: abs(s.active),
                            reverse=True):
            vprint("Processing '%s' state:\n" % state.name)
            self.process_state(state, config=config, nproc=nproc,
                               **stateargs)

    def process_state(self, state, nds=None, nproc=1,
                      config=GWSummConfigParser(), datacache=None,
                      trigcache=None, segmentcache=None, segdb_error='raise',
                      datafind_error='raise'):
        """Process data for this tab in a given state

        Parameters
        ----------
        state : `~gwsumm.state.SummaryState`
            the state to process. Can give `None` to process ALLSTATE with
            no plots, useful to load all data for other states
        nds : `bool`, optional
            `True` to use NDS to read data, otherwise read from frames.
            Use `None` to read from frames if possible, otherwise
            using NDS.
        nproc : `int`, optional
            number of parallel cores to use when reading data and making
            plots, default: ``1``
        config : `ConfigParser`, optional
            configuration for this analysis
        datacache : `~glue.lal.Cache`, optional
            `Cache` of files from which to read time-series data
        trigcache : `~glue.lal.Cache`, optional
            `Cache` of files from which to read event triggers
        segmentcache : `~glue.lal.Cache`, optional
            `Cache` of files from which to read segments
        segdb_error : `str`, optional
            if ``'raise'``: raise exceptions when the segment database
            reports exceptions, if ``'warn''`, print warnings but continue,
            otherwise ``'ignore'`` them completely and carry on.
        """
        if state:
            all_data = False
        else:
            all_data = True
            state = get_state(ALLSTATE)

        # flag those plots that were already written by this process
        for p in self.plots + self.subplots:
            if p.outputfile in globalv.WRITTEN_PLOTS:
                p.new = False

        # --------------------------------------------------------------------
        # process time-series

        # find channels that need a TimeSeries
        tschannels = self.get_channels('timeseries',
                                       all_data=all_data, read=True)
        if len(tschannels):
            vprint("    %d channels identified for TimeSeries\n"
                   % len(tschannels))
            get_timeseries_dict(tschannels, state, config=config, nds=nds,
                                nproc=nproc, cache=datacache,
                                datafind_error=datafind_error, return_=False)
            vprint("    All time-series data loaded\n")

        # find channels that need a StateVector
        svchannels = set(self.get_channels('statevector', all_data=all_data,
                                           read=True))
        odcchannels = self.get_channels('odc', all_data=all_data, read=True)
        svchannels.update(odcchannels)
        svchannels = list(svchannels)
        if len(svchannels):
            vprint("    %d channels identified as StateVectors\n"
                   % (len(svchannels) - len(odcchannels)))
            get_timeseries_dict(svchannels, state, config=config, nds=nds,
                                nproc=nproc, statevector=True,
                                cache=datacache, return_=False,
                                datafind_error=datafind_error, dtype='uint32')
            vprint("    All state-vector data loaded\n")

        # --------------------------------------------------------------------
        # process spectrograms

        # find FFT parameters
        try:
            fftparams = dict(config.nditems('fft'))
        except NoSectionError:
            fftparams = {}
        for key, val in fftparams.items():
            try:
                fftparams[key] = eval(val)
            except (NameError, SyntaxError):
                pass

        sgchannels = self.get_channels('spectrogram', 'spectrum',
                                       all_data=all_data, read=True)
        raychannels = self.get_channels('rayleigh-spectrogram',
                                        'rayleigh-spectrum',
                                        all_data=all_data, read=True)
        # for coherence spectrograms, we need all pairs of channels,
        # not just the unique ones
        csgchannels = self.get_channels('coherence-spectrogram',
                                        all_data=all_data, read=True,
                                        unique=False, state=state)

        # pad spectrogram segments to include final time bin
        specsegs = SegmentList(state.active)
        specchannels = set.union(sgchannels, raychannels, csgchannels)
        if specchannels and specsegs and specsegs[-1][1] == self.end:
            stride = max(filter(
                lambda x: x is not None,
                (get_fftparams(c, **fftparams).stride for c in specchannels),
            ))
            specsegs[-1] = Segment(specsegs[-1][0], self.end+stride)

        if len(sgchannels):
            vprint("    %d channels identified for Spectrogram\n"
                   % len(sgchannels))

            get_spectrograms(sgchannels, specsegs, config=config, nds=nds,
                             nproc=nproc, return_=False,
                             cache=datacache, datafind_error=datafind_error,
                             **fftparams)

        if len(raychannels):
            fp2 = fftparams.copy()
            fp2['method'] = fp2['format'] = 'rayleigh'
            get_spectrograms(raychannels, specsegs, config=config,
                             return_=False, nproc=nproc, **fp2)

        if len(csgchannels):
            if (len(csgchannels) % 2 != 0):
                raise ValueError("Error processing coherence spectrograms: "
                                 "you must supply exactly 2 channels for "
                                 "each spectrogram.")
            vprint("    %d channel pairs identified for Coherence "
                   "Spectrogram\n" % (len(csgchannels)/2))
            fp2 = fftparams.copy()
            fp2['method'] = 'welch'
            get_coherence_spectrograms(
                csgchannels, specsegs, config=config, nds=nds,
                nproc=nproc, return_=False, cache=datacache,
                datafind_error=datafind_error, **fp2)

        # --------------------------------------------------------------------
        # process spectra

        for channel in self.get_channels('spectrum', all_data=all_data,
                                         read=True):
            get_spectrum(channel, state, config=config, return_=False,
                         query=False, **fftparams)

        for channel in self.get_channels(
                'rayleigh-spectrum', all_data=all_data, read=True):
            fp2 = fftparams.copy()
            fp2['method'] = fp2['format'] = 'rayleigh'
            get_spectrum(channel, state, config=config, return_=False, **fp2)

        # --------------------------------------------------------------------
        # process segments

        # find flags that need a DataQualityFlag
        dqflags = set(self.get_flags('segments', all_data=all_data))
        dqflags.update(self.get_flags('timeseries', all_data=all_data,
                                      type='time-volume'))
        dqflags.update(self.get_flags('spectrogram', all_data=all_data,
                                      type='strain-time-volume'))
        if len(dqflags):
            vprint("    %d data-quality flags identified for segments\n"
                   % len(dqflags))
            get_segments(dqflags, state, config=config,
                         segdb_error=segdb_error, cache=segmentcache)

        # --------------------------------------------------------------------
        # process triggers

        for etg, channel in self.get_triggers('triggers',
                                              'trigger-timeseries',
                                              'trigger-rate',
                                              'trigger-histogram',
                                              all_data=all_data):
            get_triggers(channel, etg, state.active, config=config,
                         cache=trigcache, nproc=nproc, return_=False)

        # --------------------------------------------------------------------
        # make plots

        if all_data or self.noplots:
            vprint("    Done.\n")
            return

        # filter out plots that aren't for this state
        new_plots = [p for p in self.plots + self.subplots if p.new and
                     (p.state is None or p.state.name == state.name)]

        # separate plots into serial and parallel groups
        if int(nproc) <= 1 or not rcParams['text.usetex']:
            serial = new_plots
            parallel = []
        else:
            serial = [p for p in new_plots if not p._threadsafe]
            parallel = [p for p in new_plots if p._threadsafe]

        # process serial plots
        if serial:
            vprint("    Executing %d plots in serial:\n" % len(serial))
            multiprocess_with_queues(1, lambda p: p.process(), serial)

        # process parallel plots
        if parallel:
            nproc = min(len(parallel), nproc)
            vprint("    Executing %d plots in %d processes:\n"
                   % (len(parallel), nproc))
            multiprocess_with_queues(nproc, lambda p: p.process(), parallel)

        # record that we have written all of these plots
        globalv.WRITTEN_PLOTS.extend(p.outputfile for p in serial + parallel)

        vprint('Done.\n')

    # -------------------------------------------------------------------------
    # HTML operations

    def html_content(self, frame):
        r"""Build the #main div for this tab.

        In this construction, the <div id="id\_"> is empty, with a
        javascript hook to load the given frame into the div when ready.
        """
        page = markup.page()
        page.div(id_='main')
        page.div('', id_='content')
        page.add(str(html.load(frame, id_='content')))
        if globalv.HTML_COMMENTS_NAME:
            if globalv.IFO:
                id_ = '/%s/%s/%s' % (getpass.getuser(), globalv.IFO, self.path)
            else:
                id_ = '/%s/%s' % (getpass.getuser(), self.path)
            page.h1('Comments', class_='mt-4')
            page.add(str(html.comments_box(
                globalv.HTML_COMMENTS_NAME, identifier=id_)))
        page.div.close()
        return page

    def write_html(self, *args, **kwargs):
        writedata = kwargs.pop('writedata', True)
        vprint("Writing HTML:\n")
        for state, frame in zip(self.states, self.frames):
            idx = self.states.index(state)
            if writedata:
                self.write_state_html(state)
                vprint("    %s written\n" % frame)
            elif not os.path.isfile(self.frames[idx]):
                self.write_state_placeholder(state)
                vprint("    %s placeholder written\n" % frame)
        writehtml = kwargs.pop('writehtml', True)
        if writehtml:
            # work out whether to print comments
            comments = False
            for frame in self.frames:
                if not ('These data have not been generated yet' in
                        open(frame).read()):
                    comments = True
                    break
            if not comments:
                c = globalv.HTML_COMMENTS_NAME
                globalv.HTML_COMMENTS_NAME = None
            super(DataTab, self).write_html(*args, **kwargs)
            if not comments:
                globalv.HTML_COMMENTS_NAME = c
            vprint("    %s written\n" % self.index)

    def write_state_placeholder(self, state):
        """Write a placeholder '#main' content for this tab
        """
        email = markup.oneliner.a(
            'the DetChar group', class_='alert-link',
            href='mailto:detchar+code@ligo.org')
        page = markup.page()
        page.div(class_='row')
        page.div(class_='col-md-12')
        page.add(gwhtml.alert(
            ('Data have not been generated yet, please check back later.',
             'If this state persists for more than three or four hours, '
             'please contact %s.' % email),
            dismiss=False,
        ))
        page.div.close()
        page.div.close()

        # write to file
        idx = self.states.index(state)
        with open(self.frames[idx], 'w') as fobj:
            fobj.write(str(page))
        return self.frames[idx]

    def write_state_html(self, state):
        """Write the '#main' HTML content for this tab.

        For now, this function just links all the plots in a 2-column
        format.
        """
        page = markup.page()

        # link data
        if self.subplots:
            page.h1('Sub-plots', class_='mt-4')
            layout = get_mode() == Mode.week and [7] or [4]
            plist = [p for p in self.subplots if p.state in [state, None]]
            page.add(str(self.scaffold_plots(plots=plist, state=state,
                                             layout=layout)))

        page.div(class_='row')
        page.div(class_='col-md-12')
        channels = sorted(
            (c2 for c in self.get_channels(
                 'timeseries', 'statevector', 'spectrum', 'spectrogram', 'odc',
                 new=False) for c2 in split_channel_combination(c)),
            key=str,
        )
        if len(channels):
            page.h1('Channel information', class_='mt-4')
            headers = ['Channel', 'Type', 'Frametype', 'Sample rate', 'Units']
            data = []
            for channel in channels:
                # format CIS url and type
                if channel.frametype:
                    ftype = '<samp>%s</samp>' % channel.frametype
                else:
                    ftype = 'Unknown'
                for desc, regex in FRAMETYPE_REGEX.items():
                    if regex.match(str(channel.frametype)):
                        ftype += ' <small>[%s]</small>' % desc
                        break
                if re.search(r'\.[a-z]+\Z', channel.name):
                    name, ctype = channel.name.rsplit('.', 1)
                    c2 = get_channel(name)
                    ctype = ctype in ['rms'] and ctype.upper() or ctype.title()
                else:
                    c2 = channel
                    ctype = 'Raw'
                c = '<samp>%s</samp>' % str(channel)
                if c2.url:
                    link = markup.oneliner.a(
                        c, href=c2.url, class_='cis-link', target='_blank')
                else:
                    link = c

                # format sameple rate
                if channel.sample_rate is None:
                    rate = 'Unknown'
                elif isclose(channel.sample_rate.value, 1/60.):
                    rate = '1/60 %s' % channel.sample_rate.unit
                elif channel.sample_rate.value.is_integer():
                    rate = '{0}{1:s}'.format(int(channel.sample_rate.value),
                                             channel.sample_rate._unitstr)
                else:
                    rate = str(channel.sample_rate)
                # format unit
                if hasattr(channel, 'bits'):
                    unit = '-'
                else:
                    unit = str(channel.unit) if channel.unit else 'Unknown'
                data.append([link, ctype, ftype, rate, unit])
            page.add(str(gwhtml.table(
                headers, data, id='channel-information',
                caption="Channels used to generate data on this page")))

        allflags = sorted(set([
            (f, p) for plot in filter(
                lambda p: p.data == 'segments' and p.type != 'guardian',
                self.plots)
            for (f, p) in plot.padding.items()]), key=lambda x: x[0])
        if len(allflags):
            re_int_decimal = re.compile(r'\.00(?=(\s|\%))')
            page.h1('Segment information', class_='mt-4')
            # make summary table
            headers = ['Name', 'Defined duration [s]', 'Active duration [s]',
                       'Padding', 'Description']
            data = []
            pc = float(abs(self.span) / 100.)
            if pc.is_integer():
                pc = int(pc)
            for flag, padding in allflags:
                if padding == (0, 0):
                    padding = None
                flag = get_segments(flag, [self.span], query=False,
                                    padding={flag: padding})
                try:
                    valid = '%.2f (%.2f%%)' % (abs(flag.known),
                                               abs(flag.known) / pc)
                except ZeroDivisionError:
                    valid = '0 (0%)'
                    active = '0 (0%)'
                else:
                    active = '%.2f (%.2f%%)' % (abs(flag.active),
                                                abs(flag.active) / pc)
                valid = re_int_decimal.sub('', valid)
                active = re_int_decimal.sub('', active)
                data.append(['<samp>%s</samp>' % flag.name, valid, active,
                             padding and str(padding) or '-',
                             flag.description or ''])
            page.add(str(gwhtml.table(
                headers, data, id='segment-information',
                caption="The following flags were used in "
                        "the above data. This list does not include state "
                        "information or combinations of flags. Percentages "
                        "are calculated relative to the total duration of "
                        "%s seconds." % (pc * 100))))

            # print segment lists
            page.div(class_='mt-2', id='accordion')
            for i, (flag, padding) in enumerate(allflags):
                flag = get_segments(flag, [self.span], query=False,
                                    padding={flag: padding})
                page.div(class_='card border-info mb-1 shadow-sm')
                page.div(class_='card-header text-white bg-info')
                page.a(flag.name, class_='card-link cis-link collapsed',
                       href='#flag%d' % i, **{'data-toggle': 'collapse',
                                              'aria-expanded': 'false'})
                page.div.close()  # card-header
                page.div(id_='flag%d' % i, class_='collapse',
                         **{'data-parent': '#accordion'})
                page.div(class_='card-body')
                # write segment summary
                page.p('This flag was defined and had a known state during '
                       'the following segments:')
                page.add(str(self.print_segments(flag.known)))
                # write segment table
                page.p('This flag was active during the following segments:')
                page.add(str(self.print_segments(flag.active)))

                page.div.close()  # card-body
                page.div.close()  # collapse
                page.div.close()  # card
            page.div.close()  # accordion

        # write state information
        page.add(str(self.write_state_information(state)))

        page.div.close()
        page.div.close()

        return super(DataTab, self).write_state_html(state, plots=True,
                                                     pre=self.foreword,
                                                     post=page)

    def write_state_information(self, state):
        page = markup.page()
        # state information
        page.h1('State information', class_='mt-4')
        if state.name.lower() == ALLSTATE:
            page.p("This page was generated using all available data, "
                   "regardless of observatory operational state.")
        elif state.filename is None and state.definition is None:
            page.p("This page was generated using data in the "
                   "<strong>%s</strong> state, segments for which depend "
                   "on the input data for a given figure." % state.name)
        else:
            if state.filename:
                defn = 'via a segment file'
            elif state.MATH_DEFINITION.search(state.definition):
                defn = ('by the data condition <samp>%s</samp>'
                        % state.definition)
            else:
                defn = ('by the data-quality flag <samp>%s</samp>'
                        % state.definition)
            page.p("This page was generated using data in the "
                   "<strong>%s</strong> state. This is defined %s."
                   % (state.name, defn))
            page.add(str(self.print_segments(
                state.active, table=True,
                caption='Segments for <strong>%s</strong> state'
                        % state.name)))
        return page

    @staticmethod
    def print_segments(flag, table=False, caption=None):
        """Print the contents of a `SegmentList` in HTML
        """
        if isinstance(flag, DataQualityFlag):
            flag = flag.active
        dtype = float(abs(flag)).is_integer() and int or float
        if table:
            headers = ['GPS start', 'GPS end', 'UTC start', 'UTC end',
                       'Duration [s]']
            data = []
            for seg in flag:
                data.append([
                    dtype(seg[0]),
                    dtype(seg[1]),
                    from_gps(seg[0]).strftime('%B %d %Y %H:%M:%S.%f')[:-3],
                    from_gps(seg[1]).strftime('%B %d %Y %H:%M:%S.%f')[:-3],
                    dtype(abs(seg)),
                ])
            return gwhtml.table(headers, data, id='state-information',
                                caption=caption)
        else:
            segwizard = StringIO()
            flag.write(segwizard, format='segwizard', coltype=dtype)
            return markup.oneliner.pre(segwizard.getvalue())

    # -------------------------------------------------------------------------
    # methods

    def get_channels(self, *types, **kwargs):
        """Return the `set` of data channels required for plots of the
        given ``types``.

        Parameters
        ----------
        *types : `list` of `str`
            `list` of plot data type strings whose channel sets to return
        new : `bool`, default: `True`
            only include plots whose 'new' attribute is True

        Returns
        -------
        channels : `list`
            an alphabetically-sorted `list` of channels
        """
        isnew = kwargs.pop('new', True)
        if kwargs.pop('unique', True):
            out = set()
            add = out.update
        else:
            out = list()
            add = out.extend
        for plot in self.plots:
            if plot.data not in types:
                continue
            if isnew and not plot.new:
                continue
            skip = False
            for key, val in kwargs.items():
                if getattr(plot, key) != val:
                    skip = True
                    break
            if skip:
                continue
            add(plot.channels)
            if plot.data == 'odc':
                add(plot.get_bitmask_channels())
        return out

    def get_flags(self, *types, **kwargs):
        """Return the `set` of data-quality flags required for plots of the
        given ``types``.

        Parameters
        ----------
        *types : `list` of `str`
            `list` of plot type strings whose flag sets to return

        Returns
        -------
        flags : `list`
            an alphabetically-sorted `list` of flags
        """
        isnew = kwargs.pop('new', True)
        uniq = kwargs.pop('unique', True)
        out = set()
        for plot in self.plots:
            if plot.data not in types:
                continue
            if isnew and not plot.new:
                continue
            skip = False
            for key, val in kwargs.items():
                if getattr(plot, key) != val:
                    skip = True
                    break
            if skip:
                continue
            if uniq:
                out.update([f for cflag in plot.flags for f in
                            re_flagdiv.split(cflag)[::2] if f])
            else:
                out.update(plot.flags)
        return sorted(out, key=lambda dqf: str(dqf))

    def get_triggers(self, *types, **kwargs):
        """Return the `set` of data-quality flags required for plots of the
        given ``types``.

        Parameters
        ----------
        *types : `list` of `str`
            `list` of plot type strings whose flag sets to return

        Returns
        -------
        flags : `list`
            an alphabetically-sorted `list` of flags
        """
        isnew = kwargs.pop('new', True)
        out = set()
        for plot in self.plots:
            if plot.type not in types:
                continue
            if isnew and not plot.new:
                continue
            skip = False
            for key, val in kwargs.items():
                if getattr(plot, key) != val:
                    skip = True
                    break
            if skip:
                continue
            for channel in plot.channels:
                out.add((plot.etg, channel))
        return sorted(out, key=lambda ch: ch[1].name)


register_tab(DataTab)
register_tab(DataTab, name='default')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""
"""

import re
import abc
import os.path
from io import StringIO

from lxml import etree

from gwdetchar.plot import texify

re_bit_label = re.compile(r'\[(?P<idx>.*)\] (?P<label>.*)')
re_source_label = re.compile(r'(?P<label>.*) \[(?P<flag>.*)\]')

HOVERSCRIPT = """
<script type="text/ecmascript">
<![CDATA[    function init(evt) {
      if ( window.svgDocument == null ) {
        svgDocument = evt.target.ownerDocument;
      }
      var labels = svgDocument.getElementsByClassName('mpl-label');
      for (var i=0; labels[i]; i++) {
        labels[i].setAttribute('visibility', 'hidden');
      }
    }

    function ShowLabel(obj) {
      var cur = obj.id.substr(obj.id.lastIndexOf('_')+1);
      var tip = svgDocument.getElementById('label_' + cur);
      tip.setAttribute('visibility',"visible")
    }

    function HideLabel(obj) {
      var cur = obj.id.substr(obj.id.lastIndexOf('_')+1);
      var tip = svgDocument.getElementById('label_' + cur);
      tip.setAttribute('visibility',"hidden")
    }]]>
  </script>

"""

HTML_WRAPPER = """
<html>
<body>
  <style>
    body { margin: 0 !important }
  </style>
</head>
<div>
<object type="image/svg+xml" data="%s"
Your browser cannot display this SVG
</object>
</div>
</body>
</html>
"""


class SvgMixin(object, metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwargs):
        super(SvgMixin, self).__init__(*args, **kwargs)
        self.preview_labels = False

    def finalize(self, outputfile=None, close=True, **savekwargs):
        if outputfile is None:
            outputfile = self.outputfile
        # make SVG
        if outputfile.endswith('.svg'):
            self.draw_svg(outputfile)
        # or continue as normal
        else:
            return super(SvgMixin, self).finalize(
                outputfile=outputfile, close=close, **savekwargs)

    @abc.abstractmethod
    def draw_svg(self, outputfile):
        pass

    def finalize_svg(self, tree, outputfile, script=None):
        if script:
            tree.insert(0, etree.XML(script))
        etree.ElementTree(tree).write(outputfile)
        # write HTML wrapper
        html = self.outputfile.replace('.svg', '.html')
        with open(html, 'w') as f:
            f.write(HTML_WRAPPER % os.path.basename(self.outputfile))
        return html


class DataLabelSvgMixin(SvgMixin):
    def draw_svg(self, outputfile):
        for ax in self.plot.axes:
            for line in ax.lines:
                line.set_rasterized(True)

        # render image
        super(SvgMixin, self).finalize(
            outputfile=outputfile.replace('.svg', '.png'), close=False)

        # make new text labels for the channel names
        ax = self.plot.axes[0]
        leg = ax.legend_
        texts = []
        if leg is not None:
            for i, (text, line) in enumerate(
                    zip(leg.get_texts(), leg.get_lines())):
                try:
                    label, source = re_source_label.match(
                        text.get_text()).groups()
                except (AttributeError, ValueError):
                    continue
                channel = texify(str(source))
                text.set_text(label)
                t2 = ax.text(
                    0.994, 1.02, channel, ha='right', va='bottom',
                    fontsize=text.get_fontsize(), zorder=1000,
                    transform=ax.transAxes,
                    bbox={'facecolor': 'white', 'edgecolor': 'lightgray',
                          'pad': 10.})
                text.set_gid('leg_text_%d' % i)
                line.set_gid('leg_patch_%d' % i)
                t2.set_gid('label_%d' % i)
                texts.append(t2)

        # tmp save
        f = StringIO()
        self.plot.save(f, format='svg')

        # parse svg
        tree, xmlid = etree.XMLID(f.getvalue())
        tree.set('onload', 'init(evt)')

        # add effects
        for i in range(len(texts)):
            pl = xmlid['leg_patch_%d' % i]
            ll = xmlid['leg_text_%d' % i]
            tl = xmlid['label_%d' % i]
            pl.set('cursor', 'pointer')
            pl.set('onmouseover', "ShowLabel(this)")
            pl.set('onmouseout', "HideLabel(this)")
            ll.set('cursor', 'pointer')
            ll.set('onmouseover', "ShowLabel(this)")
            ll.set('onmouseout', "HideLabel(this)")
            tl.set('class', 'mpl-label')
            tl.set('visibility', 'hidden')

        return self.finalize_svg(tree, outputfile, script=HOVERSCRIPT)


class SegmentLabelSvgMixin(SvgMixin):
    def draw_svg(self, outputfile):
        # render image
        super(SvgMixin, self).finalize(outputfile=StringIO(), close=False)
        ax = self.plot.axes[0]
        collections = [c for c in ax.collections if hasattr(c, '_ignore')]
        # reset labels
        labels = {}
        for i, t in enumerate(ax.get_yaxis().get_ticklabels()):
            text = t.get_text()
            if not text:
                continue
            x, y = t.get_position()
            m1 = re_bit_label.match(text)
            m2 = re_source_label.match(text)
            if m1:
                idx, label = m1.groups()
                t2 = ax.text(0.01, y, label, ha='left',
                             fontsize=t.get_fontsize(), va='center',
                             transform=t.get_transform())
                t2.set_bbox({'alpha': 0.5, 'facecolor': 'white',
                             'edgecolor': 'none'})
                t.set_text(idx)
                t.set_bbox(None)
                t.set_position((0, y))
                t.set_ha('right')
                t.set_fontsize('14')
                labels[text] = (idx, t2)
            elif m2:
                label, flag = m2.groups()
                t2 = ax.text(x, y, text, ha='left', bbox=t._bbox.copy(),
                             fontsize=t.get_fontsize(), va='center',
                             transform=t.get_transform())
                t.set_text(label)
                labels[text] = (label, t2)
        ax._insetlabels = None

        j = 0
        for i, collection in enumerate(collections):
            try:
                idx, tickl = labels[collection.get_label()]
            except KeyError:
                continue
            else:
                labels[collection] = tickl
            collection.set_label(idx)
            if not tickl.get_gid():
                tickl.set_gid('label_%d' % j)
                j += 1
            collection.set_gid('collection_%d_%s' % (i, tickl.get_gid()))

        # render as and parse SVG
        f = StringIO()
        self.plot.save(f, format='svg')
        tree, xmlid = etree.XMLID(f.getvalue())
        tree.set('onload', 'init(evt)')
        # set mouse events for visible labels
        for i, collection in enumerate(collections):
            try:
                tickl = labels[collection]
            except KeyError:
                continue
            cel = xmlid[collection.get_gid()]
            cel.set('cursor', 'pointer')
            cel.set('onmouseover', "ShowLabel(this)")
            cel.set('onmouseout', "HideLabel(this)")
            tel = xmlid[tickl.get_gid()]
            tel.set('class', 'mpl-label')
            if not self.preview_labels:
                tel.set('visibility', 'hidden')

        # ignore hover on text events
        for key in xmlid:
            if key.startswith('text_') or key.startswith('label_'):
                xmlid[key].set('style', 'pointer-events: none;')

        return self.finalize_svg(tree, outputfile, script=HOVERSCRIPT)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.


"""Set of global memory variables for GWSumm package
"""

import time

from gwpy.time import to_gps
from gwpy.segments import DataQualityDict
from gwpy.detector import ChannelList

CHANNELS = ChannelList()
STATES = {}

DATA = {}
SPECTROGRAMS = {}
SPECTRUM = {}
COHERENCE_COMPONENTS = {}
COHERENCE_SPECTRUM = {}
SEGMENTS = DataQualityDict()
TRIGGERS = {}

VERBOSE = False
PROFILE = False
START = time.time()

# run time variables
MODE = 0
WRITTEN_PLOTS = []
NOW = int(to_gps('now'))
HTMLONLY = False

# comments
IFO = None
HTML_COMMENTS_NAME = None



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Definition of the `SummaryState` class.
"""

import datetime
import re
import operator
from configparser import (NoOptionError, DEFAULTSECT)

from astropy.time import Time

from gwpy.detector import get_timezone_offset
from gwpy.segments import (Segment, SegmentList, DataQualityFlag)
from gwpy.time import (to_gps, from_gps)

from .. import globalv
from ..config import (GWSummConfigParser)
from ..utils import re_cchar
from ..segments import get_segments
from ..data import get_timeseries

MATHOPS = {
    '<': operator.lt,
    '<=': operator.le,
    '=': operator.eq,
    '>=': operator.ge,
    '>': operator.gt,
    '==': operator.is_,
    '!=': operator.is_not,
}


class SummaryState(DataQualityFlag):
    """An operating state over which to process a `~gwsumm.tabs.DataTab`.

    Parameters
    ----------
    name : `str`
        name for this state
    known : `~gwpy.segments.SegmentList`, optional
        list of known segments
    active : `~gwpy.segments.SegmentList`, optional
        list of active segments
    description : `str`, optional
        text describing what this state means
    definition : `str`, optional
        logical combination of flags that define known and active segments
        for this state (see :attr:`documentation <SummaryState.definition>`
        for details)
    key : `str`, optional
        registry key for this state, defaults to :attr:`~SummaryState.name`
    """
    MATH_DEFINITION = re.compile(r'(%s)' % '|'.join(MATHOPS.keys()))

    def __init__(self, name, known=SegmentList(), active=SegmentList(),
                 description=None, definition=None, hours=None, key=None,
                 filename=None, url=None):
        """Initialise a new `SummaryState`
        """
        # allow users to specify known as (start, end)
        if (isinstance(known, Segment) or
                (isinstance(known, tuple) and len(known) == 2 and
                 not isinstance(known[0], tuple))):
            known = [known]
        super(SummaryState, self).__init__(name=name, known=known,
                                           active=active)
        self.description = description
        if definition:
            self.definition = re.sub(r'(\s|\n)', '', definition)
        else:
            self.definition = None
        self.key = key
        self.hours = hours
        self.url = url
        if known and active:
            self.ready = True
        else:
            self.ready = False
        self.filename = filename

    @property
    def name(self):
        """Name of this state
        """
        return self._name

    @name.setter
    def name(self, n):
        self._name = n

    @property
    def tag(self):
        """File tag for images generated using this state
        """
        return re_cchar.sub("_", self.name).upper()

    @property
    def start(self):
        """GPS start time of this state's validity.
        """
        return self.extent[0]

    @property
    def end(self):
        """GPS end time of this state's validity
        """
        return self.extent[1]

    @property
    def definition(self):
        r"""The combination of data-quality flags that define this
        `SummaryState`

        For example::

           >>> state = SummaryState(definition='L1:DMT-SCIENCE:1')

        would define a `SummaryState` based on the validity and activity of
        the single flag ``'L1:DMT-SCIENCE:1'`` (the science-mode flag for the
        LIGO Livingston Observatory interferometer).
        Similarly::

           >>> state = SummaryState(
                   definition='L1:DMT-SCIENCE:1&!L1:DMT-LIGHTDIP_10_PERCENT:1')

        would define a `SummaryState` as active when the ``'L1:DMT-SCIENCE:1'``
        flag was active and the ``'L1:DMT-LIGHTDIP_10_PERCENT:1'`` flag was
        not active.

        The following logical identifiers are acceptable:

        ==  ===================================================================
        &   Union (i.e. flag1 **and** flag2 must be active)
        \|  Intersection (i.e. flag1 **or** flag2 must be active)
        &!  One-sided difference (i.e. flag1 is active and flag2 **is not**\
            active)
        !=  Two-sided difference (i.e. flag1 is active and flag2 is not **OR**\
            flag2 is active and flag2 is not)
        ==  ===================================================================

        :type: str
        """
        return self._definition

    @definition.setter
    def definition(self, d):
        if d:
            self._definition = str(d)
        else:
            self._definition = None

    @property
    def key(self):
        """The registry key for this `SummaryState`.

        :type: `str`
        """
        if self._key is not None:
            return self._key
        else:
            return self._name

    @key.setter
    def key(self, k):
        self._key = k

    @classmethod
    def from_ini(cls, config, section):
        """Create a new `SummaryState` from a section in a `ConfigParser`.

        Parameters
        ----------
        config : :class:`~gwsumm.config.GWConfigParser`
            customised configuration parser containing given section
        section : `str`
            name of section to parse

        Returns
        -------
        `SummaryState`
            a new state, with attributes set from the options in the
            configuration
        """
        config = GWSummConfigParser.from_configparser(config)
        # get span times
        start = config.getint(section, 'gps-start-time')
        end = min(globalv.NOW, config.getint(section, 'gps-end-time'))
        # get parameters
        params = dict(config.nditems(section))
        # parse name
        name = params.pop('name', section)
        if re.match(r'state[-\s]', name):
            name = section[6:]
        # get hours
        hours = params.pop('hours', None)
        if hours is not None:
            segs = re.split(r'(,|, )', hours)[::2]
            hours = []
            offset = 0
            for seg in segs:
                try:
                    # parse hour segment
                    hours.append(list(map(float, seg.split('-', 1))))
                except ValueError:
                    # parse time-zone
                    if seg == segs[-1]:
                        if seg.lower() == 'utc':
                            offset = 0
                        elif seg.lower() == 'local':
                            try:
                                ifo = config.get(DEFAULTSECT, 'ifo')
                            except NoOptionError:
                                raise ValueError("The relevant IFO must be "
                                                 "given either from the --ifo "
                                                 "command-line option, or the "
                                                 "[DEFAULT] section of any "
                                                 "INI file")
                            offset = get_timezone_offset(ifo, from_gps(start))
                        else:
                            offset = get_timezone_offset(seg, from_gps(start))
                    else:
                        raise
            # apply time-zone
            for i, (h0, h1) in enumerate(hours):
                hours[i] = (h0 - offset / 3600., h1 - offset / 3600.)
        # generate state
        return cls(name, known=[(start, end)], hours=hours, **params)

    def _fetch_segments(self, config=GWSummConfigParser(), **kwargs):
        kwargs.setdefault('url', self.url)
        segs = get_segments([self.definition], self.known, config=config,
                            **kwargs)[self.definition].round(contract=True)
        self.known = segs.known
        self.active = segs.active
        return self

    def _fetch_data(self, channel, thresh, op, config=GWSummConfigParser(),
                    **kwargs):
        # if 0 is out-of-state, allowing padding gaps with 0.
        if (
                (op == '<' and thresh <= 0.) or
                (op == '<=' and thresh < 0.) or
                (op == '>' and thresh >= 0.) or
                (op == '>=' and thresh > 0.) or
                (op in ('=', '==') and thresh != 0.) or
                (op == '!=' and thresh == 0.)
        ):
            kwargs.setdefault('pad', 0.)
        data = get_timeseries(channel, self.known, config=config, **kwargs)
        for ts in data:
            if isinstance(thresh, (float, int)) and ts.unit is not None:
                thresh *= ts.unit
            segs = MATHOPS[op](ts, thresh).to_dqflag()
            try:
                globalv.SEGMENTS[self.definition] += segs
            except KeyError:
                globalv.SEGMENTS[self.definition] = segs
        return self._fetch_segments(query=False)

    def _read_segments(self, filename):
        segs = DataQualityFlag.read(filename, self.definition)
        # XXX HACK around malformed segment files with no segment_summary table
        if segs.active and not segs.known:
            segs.known = type(segs.active)(segs.active)
        if self.known:
            self.known = self.known & segs.known
            self.active = self.known & segs.active
        else:
            self.known = segs.known
            self.active = segs.active
        return self

    def fetch(self, config=GWSummConfigParser(),
              segmentcache=None, segdb_error='raise',
              datacache=None, datafind_error='raise', nproc=1, nds=None,
              **kwargs):
        """Finalise this state by fetching its defining segments,
        either from global memory, or from the segment database
        """
        # check we haven't done this before
        if self.ready:
            return self
        # fetch data
        match = self.MATH_DEFINITION.search(str(self.definition))
        if self.filename:
            self._read_segments(self.filename)
        elif self.definition and match is not None:
            channel, thresh = self.definition.split(match.groups()[0])
            channel = channel.rstrip()
            thresh = float(thresh.strip())
            self._fetch_data(channel, thresh, match.groups()[0], config=config,
                             cache=datacache, nproc=nproc, nds=nds,
                             datafind_error=datafind_error, **kwargs)
        # fetch segments
        elif self.definition:
            self._fetch_segments(config=config, cache=segmentcache,
                                 segdb_error=segdb_error, **kwargs)
        # fetch null
        else:
            start = config.getfloat(DEFAULTSECT, 'gps-start-time')
            end = config.getfloat(DEFAULTSECT, 'gps-end-time')
            self.known = [(start, end)]
            self.active = self.known
        # restrict to given hours
        if self.hours:
            segs_ = SegmentList()
            # get start day
            d = Time(float(self.start), format='gps', scale='utc').datetime
            d = d.replace(hour=0, minute=0, second=0, microsecond=0)
            end_ = Time(float(self.end), format='gps', scale='utc').datetime
            while d < end_:
                # get GPS of day
                t = to_gps(d)
                # for each [start, end) hour pair, build a segment
                for h0, h1 in self.hours:
                    segs_.append(Segment(t + h0 * 3600, t + h1*3600))
                # increment and return
                d += datetime.timedelta(1)
            self.known &= segs_
            self.active &= segs_
        # FIXME
        self.ready = True
        return self

    def copy(self):
        new = super(SummaryState, self).copy()
        new.description = self.description
        new.definition = self.definition
        new.ready = self.ready
        return new
    copy.__doc__ = DataQualityFlag.copy.__doc__

    def __str__(self):
        return self.name



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Registry for GWSumm data tabs.

All Tabs should be registered for easy identification from the
configuration INI files
"""

from ..utils import re_quote

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

__all__ = ['register_tab', 'get_tab']

_TABS = {}


def register_tab(tab, name=None, force=False):
    """Register a new summary `Tab` to the given ``name``

    Parameters
    ----------
    tab : `type`
        defining Class for this tab type.
    name : `str`, optional
        unique descriptive name for this type of tab, must not
        contain any spaces, e.g. 'hveto'. If ``name=None``, the `Tab.type`
        class attribute of the given tab will be used.
    force : `bool`
        overwrite existing registration for this type

    Raises
    ------
    ValueError
        if name is already registered and ``force`` not given as `True`
    """
    if name is None:
        name = tab.type
    if name not in _TABS or force:
        _TABS[name] = tab
    else:
        raise ValueError("Tab '%s' has already been registered to the %s "
                         "class" % (name, _TABS[name].__name__))


def get_tab(name):
    """Query the registry for the tab class registered to the given
    name
    """
    name = re_quote.sub('', name)
    try:
        return _TABS[name]
    except KeyError:
        raise ValueError("No Tab registered with name '%s'" % name)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Extensions to the spectrum plot for noise budgets
"""

import re

import numpy

from gwpy.segments import SegmentList

from gwdetchar.plot import texify

from ..data import get_spectrum
from .registry import (get_plot, register_plot)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


class NoiseBudgetPlot(get_plot('spectrum')):
    """Plot of a noise budget ASD
    """
    type = 'noise-budget'
    data = 'spectrum'
    defaults = get_plot('spectrum').defaults.copy()
    defaults.update({
        'xscale': 'log',
        'yscale': 'log',
        'format': 'asd',
        'sum-label': 'Sum of noises',
        'sum-linestyle': '--',
        'sum-color': 'black',
        'residual-label': 'Residual',
        'residual-linestyle': ':',
        'residual-color': 'grey',
    })

    def _parse_extra_params(self, prefix, **defaults):
        """Parse parameters for an extra plot element
        """
        re_prefix = re.compile(r'\A%s[-_]' % prefix.rstrip('-_'))
        extras = defaults.copy()
        for key in list(self.pargs):
            m = re_prefix.match(key)
            if m:
                extras[key[m.span()[1]:]] = self.pargs.pop(key)
        return extras

    def parse_sum_params(self, **defaults):
        return self._parse_extra_params('sum', **defaults)

    def parse_residual_params(self, **defaults):
        return self._parse_extra_params('residual', **defaults)

    def _draw(self):
        """Load all data, and generate this `SpectrumDataPlot`
        """
        plot = self.init_plot()
        ax = plot.gca()
        ax.grid(b=True, axis='both', which='both')

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        texify(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')

        # get spectrum format: 'amplitude' or 'power'
        sdform = self.pargs.pop('format')

        # parse plotting arguments
        plotargs = self.parse_plot_kwargs()
        legendargs = self.parse_legend_kwargs()

        # add data
        sumdata = []
        for i, (channel, pargs) in enumerate(zip(self.channels, plotargs)):
            if self.state and not self.all_data:
                valid = self.state
            else:
                valid = SegmentList([self.span])

            data = get_spectrum(str(channel), valid, query=False,
                                format=sdform, method=None)[0]
            if i and data.size:
                sumdata.append(data)
            else:
                darmdata = data

            # anticipate log problems
            if self.logx:
                data = data[1:]
            if self.logy:
                data.value[data.value == 0] = 1e-100

            pargs.setdefault('zorder', -i)
            ax.plot(data, **pargs)

        # assert all noise terms have the same resolution
        if any([x.dx != sumdata[0].dx for x in sumdata]):
            raise RuntimeError("Noise components have different resolutions, "
                               "cannot construct sum of noises")
        # reshape noises if required
        n = max(x.size for x in sumdata)
        for i, d in enumerate(sumdata):
            if d.size < n:
                sumdata[i] = numpy.resize(
                    numpy.require(d, requirements=['O']), (n,))

        # plot sum of noises
        sumargs = self.parse_sum_params()
        sum_ = numpy.zeros_like(darmdata ** 2, shape=(n,))
        for d in sumdata:
            d.override_unit(darmdata.unit)  # enforce consistent units
            sum_ += d ** 2
        ax.plot(sum_ ** (1/2.), zorder=1, **sumargs)
        ax.lines.insert(1, ax.lines.pop(-1))

        # plot residual of noises
        if not self.pargs.pop('no-residual', False):
            resargs = self.parse_residual_params()
            try:
                residual = (darmdata ** 2 - sum_) ** (1/2.)
            except ValueError:
                if not darmdata.size:  # if no data, just copy nothing
                    residual = darmdata
                else:  # other error
                    raise
            ax.plot(residual, zorder=-1000, **resargs)
            ax.lines.insert(1, ax.lines.pop(-1))

        # finalize
        self.apply_parameters(ax, **self.pargs)
        ax.legend(**legendargs)

        return self.finalize()


register_plot(NoiseBudgetPlot)


class RelativeNoiseBudgetPlot(get_plot('spectrum')):
    """Spectrum plot for a `SummaryTab`
    """
    type = 'noise-budget-ratio'
    data = 'spectrum'
    defaults = get_plot('spectrum').defaults.copy()
    defaults.update({
        'xscale': 'log',
        'yscale': 'log',
        'format': 'asd',
    })

    def _draw(self):
        """Load all data, and generate this `SpectrumDataPlot`
        """
        plot = self.init_plot()
        ax = plot.gca()
        ax.grid(b=True, axis='both', which='both')

        if self.state:
            self.pargs.setdefault(
                'suptitle',
                '[%s-%s, state: %s]' % (self.span[0], self.span[1],
                                        texify(str(self.state))))
        suptitle = self.pargs.pop('suptitle', None)
        if suptitle:
            plot.suptitle(suptitle, y=0.993, va='top')

        # get spectrum format: 'amplitude' or 'power'
        sdform = self.pargs.pop('format')

        # parse plotting arguments
        plotargs = self.parse_plot_kwargs()[0]

        # add data
        sumdata = []
        for i, channel in enumerate(self.channels):
            if self.state and not self.all_data:
                valid = self.state
            else:
                valid = SegmentList([self.span])

            data = get_spectrum(str(channel), valid, query=False,
                                format=sdform, method=None)[0]
            if i and data.size:
                sumdata.append(data)
            else:
                target = data

        if target.size:
            # assert all noise terms have the same resolution
            if any([x.dx != target.dx for x in sumdata]):
                raise RuntimeError("Noise components have different "
                                   "resolutions, cannot construct "
                                   "sum of noises")
            # reshape noises if required
            n = target.size
            for i, d in enumerate(sumdata):
                if d.size < n:
                    sumdata[i] = numpy.resize(
                        numpy.require(d, requirements=['O']), (n,))

            # calculate sum of noises
            sum_ = numpy.zeros_like(sumdata[0] ** 2, shape=(n,))
            for d in sumdata:
                d.override_unit(sumdata[0].unit)  # enforce consistent units
                sum_ += d ** 2
            sum_ **= (1/2.)

            relative = sum_ / target
        else:  # no data, so just use anything as a proxy
            relative = target

        # plot ratio of h(t) to sum of noises
        ax.plot(relative, **plotargs)

        # finalize plot
        self.apply_parameters(ax, **self.pargs)

        return self.finalize()


register_plot(RelativeNoiseBudgetPlot)



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for gwsumm.html.html5
"""

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

import os
import shutil

from markdown import markdown

from gwdetchar.utils import parse_html

from .. import html5

# global variables

URL = 'https://github.com/gwpy/gwsumm'
LOAD = """<script>
    jQuery.ajax({
        url : '%s',
        type : 'GET',
        success: function(data, statusText, jqhxr){%s},
        error: function(xhr, status, error){%s}
        });
</script>"""

BOX = """<div id="disqus_thread">
<script type="text/javascript">
    var disqus_shortname = "Test";
    var disqus_identifier = "test";
    var disqus_title = "Test";
    var disqus_url = "%s";

        (function() {
            var dsq = document.createElement('script');
            dsq.type = 'text/javascript'; dsq.async = true;
            dsq.src = '//' + disqus_shortname + '.disqus.com/embed.js';
            (document.getElementsByTagName('head')[0] ||
             document.getElementsByTagName('body')[0]).appendChild(dsq);
        })();

</script>
<noscript>Please enable JavaScript to view the</noscript>
<a href="https://disqus.com/?ref_noscript">comments powered by Disqus</a>
</div>"""

CONTENTS = """Heading
=======

This is a test.

* Bullet 1
* Bullet 2"""

DIALOG = ('<button title="Test" id="id-btn" class="btn-float btn-open shadow" '
          'data-id="#id">T</button>\n<div class="dialog" title="Test" '
          'id="id">\n%s\n</div>') % markdown(CONTENTS)

OVERLAY = (
    '<button title="Overlay figures" id="overlay-btn" class="btn-float '
    'btn-open shadow" data-id="#overlay"><i class="fas fa-layer-group"></i>'
    '</button>\n<div class="dialog" title="Overlay figures" id="overlay">\n'
    '<h1>Overlay figures for easy comparison</h1>\n<hr class="row-divider" />'
    '\n<div class="row" id="overlay-outer">\n<div class="col-md-4">\n'
    '<div class="card card-body shadow-sm" id="overlay-info">\n'
    '<h4>Instructions</h4>\n%s\n</div>\n</div>\n<div class="col-md-8">\n'
    '<div class="text-center">\n<a title="Overlay all selected figures" '
    'class="btn btn-light shadow-sm" id="overlay-figures">Overlay</a>\n'
    '<a title="Download overlay figure" class="btn btn-light shadow-sm" '
    'id="download-overlay">Download</a>\n<a title="Clear all figure '
    'selections" class="btn btn-light shadow-sm" id="clear-figures">Clear</a>'
    '\n</div>\n<br />\n<canvas id="overlay-canvas" />\n</div>\n</div>\n'
    '</div>') % markdown(html5.OVERLAY_INSTRUCTIONS)


# test utilities

def test_expand_path():
    assert html5._expand_path(URL) == URL


def test_load_state():
    state = html5.load_state('test')
    assert parse_html(str(state)) == parse_html(
        '<script>\nif (location.hash.length <= 1) {\n'
        '    jQuery("#state_test").load_state("test");\n}\n'
        '</script>')


def test_load():
    # test local url
    content = html5.load('test')
    assert parse_html(content) == parse_html(
        '<script>jQuery("#main").load("test");</script>')
    # test non-local url
    success = 'jQuery(\"#main\").html(data);'
    errormsg = ('alert(\"Cannot load content from %r, use '
                'browser console to inspect failure.\");' % URL)
    content = html5.load(URL)
    assert parse_html(content) == parse_html(LOAD % (URL, success, errormsg))
    # test with non-string error argument
    success = 'jQuery(\"#main\").html(data);'
    error = 'Failed to load content from %r' % URL
    errormsg = ('jQuery(\"#main\").html(\"<div class=\'alert alert-warning '
                'text-justify shadow-sm\'><p>%s</p></div>\");' % error)
    content = html5.load(URL, error=1)
    assert parse_html(content) == parse_html(LOAD % (URL, success, errormsg))


def test_load_custom():
    error = 'Error'
    success = 'document.write(\"Success\")'
    errormsg = ('jQuery(\"#main\").html(\"<div class=\'alert alert-warning '
                'text-justify shadow-sm\'><p>%s</p></div>\");' % error)
    # test local url
    content = html5.load('test', success=success, error=error)
    assert parse_html(content) == parse_html(
        LOAD % ('test', success, errormsg))
    # test non-local url
    content = html5.load(URL, success=success, error=error)
    assert parse_html(content) == parse_html(LOAD % (URL, success, errormsg))


def test_comments_box():
    box = html5.comments_box('Test', identifier='test', title='Test', url=URL)
    assert parse_html(str(box)) == parse_html(BOX % URL)


def test_ldvw_qscan_single():
    button = html5.ldvw_qscan('X1:TEST', 0)
    assert parse_html(str(button)) == parse_html(
        '<a href="https://ldvw.ligo.caltech.edu/ldvw/view?act=doplot&amp;'
        'chanName=X1:TEST&amp;doQxfrm=yes&amp;strtTime=0&amp;&amp;'
        'qxfrm_pltfrq=10 inf&amp;qxfrm_srchqrng=4 100&amp;'
        'qxfrm_plttimes=0.5 2 8" target="_blank" rel="external" '
        'class="btn btn-outline-secondary btn-sm" title="Q-scan X1:TEST at 0 '
        'via LDVW">Q-scan</a>')


def test_ldvw_qscan_batch():
    button = html5.ldvw_qscan('X1:TEST', (0,))
    assert parse_html(str(button)) == parse_html(
        '<a href="https://ldvw.ligo.caltech.edu/ldvw/Wdq?submitAct=go&amp;'
        'wdq_ifo=X1&amp;wdq_cmap=viridis&amp;wdq_gps=0&amp;wdq_prog=py-Omega&'
        'amp;goBtn=goBtn" target="_blank" rel="external" class="btn '
        'btn-outline-secondary btn-sm" title="Batch-process omega scans of '
        'the loudest triggers via LDVW">Launch omega scans</a>')


def test_dialog_box(tmpdir):
    mdfile = os.path.join(str(tmpdir), 'test.md')
    with open(mdfile, 'w') as f:
        f.write(CONTENTS)
    box = html5.dialog_box(mdfile, 'Test', 'id', 'T')
    assert parse_html(str(box)) == parse_html(DIALOG)
    shutil.rmtree(str(tmpdir), ignore_errors=True)


def test_overlay_canvas():
    box = html5.overlay_canvas()
    assert parse_html(str(box)) == parse_html(OVERLAY)
